import re
import unittest
import numbers
import sys

if sys.version_info[0] < 3:
    from operator import div  # Python 2
else:
    from operator import truediv as div

import numpy

from GTC import *
from GTC.context import _context 
from GTC.reporting import *
from GTC.lib import (
    UncertainComplex,
    willink_hall,
    _is_uncertain_complex_constant
)

from testing_tools import *

TOL = 1E-13 
LOG10_E = cmath.log10(cmath.e)

numpy.seterr(all='raise')

#----------------------------------------------------------------------------
def is_positive_definite(x):
    """
    Test whether 'x' is positive definite, or, if one of the
    diagonal elements is zero, then the off-diagonal elements
    must also be zero.
    
    """
    
    if (x[0,0] == 0.0) or (x[1,1] == 0.0):
        # If one diagonal element is zero, both
        # off-diagonal elements must be zero.
        return (x[1,0] == 0.0) and (x[0,1] == 0.0)
    elif abs(x[1,0]-x[0,1]) > 1E-16:
        # Numpy.linalg.cholesky does not test matrix symmetry!!
        return False
    else:
        try:
            numpy.linalg.cholesky(x)
            return True
        except numpy.linalg.LinAlgError:
            return False

#----------------------------------------------------------------------------
def to_std_uncertainty(x):
    """
    Return a 2x2 std_uncertainty matrix
    and a correlation coefficient 'r' from x (a real number,
    a 2-element numeric sequence, or a 4-element sequence)
    
    """
    d = numpy.size(x)
    if( d == 1 ):
        return ( numpy.asarray( [[x,0.],[0.,x]] ), 0.0 )
    elif( d == 2 ):
        # Assume that these are the diagonal standard uncertainies
        x0,x1 = x
        return ( numpy.asarray( [[x0,0.],[0.,x1]] ), 0.0 )
    elif ( d == 4 ):       
        tmp = numpy.asarray( x, float )
        tmp.shape = ( 2, 2 )
        assert is_positive_definite( tmp ),"not +ve definite: '%s'"
        
        u1 = math.sqrt(tmp[0,0])
        u2 = math.sqrt(tmp[1,1])
        return ( numpy.asarray( [[u1,0.],[0.,u2]] ), tmp[0,1] /(u1*u2) )
    else:
        raise TypeError('cast failed')

#----------------------------------------------------------------------------
def number_to_matrix(*args):
    """
    Create a (2,2N) numpy.ndarray from a series of complex numbers.
    
    """
    mat = numpy.empty( (2,2*len(args)),float )
    for i,a in enumerate(args):
        c = complex( a )
        mat[:,2*i:2*i+2] = ( (c.real, -c.imag) , (c.imag, c.real) ) 

    return mat

#----------------------------------------------------------------------------
def array_to_sequence(m):
    """
    -> a sequence from a numpy array or matrix
    
    A 2-by-2 matrix [ (1,2), (3,4) ] is returned [ 1,2,3,4 ]
    
    """
    return list(m.flat)

#----------------------------------------------------------------------------
class NamesUncertainComplex(unittest.TestCase):

    def test_is_ucomplex(self):
        x = ureal(1,1)
        y = 1+3j
        z = ucomplex(1,1)
        
        self.assertTrue( not is_ucomplex(x) )
        self.assertTrue( not is_ucomplex(y) )
        self.assertTrue( is_ucomplex(z) )
    
#----------------------------------------------------------------------------
class NamesComplex(unittest.TestCase):
    """
    - labels can be assigned to any elementary UN at construction
    - labels can be assigned to intermediate UNs by a special naming function
    - when a label is assigned to a complex quantity, labels are automatically
      assigned to the real and imaginary components, of the form 'label_re' and 'label_im'.
    - the context maintains a register mapping uids to labels of elementary UNs
    - when a UN is garbage-collected, the register entry is removed
    - labels are read by a special function that will return the uid of an elementary if
      no label has been assigned, or a generic 'anon' otherwise
    """
    def setUp(self):
        self._context = _context

        z1 = ucomplex(1,(1,1),label='z1')
        self.assertEqual(label(z1),'z1')
        self.assertEqual(label(z1.real),'z1_re')
        self.assertEqual(label(z1.imag),'z1_im')
        
        z2 = ucomplex(1,(1,1))
        self.assertEqual(label(z2),None)
        self.assertEqual(label(z2.real),None)
        self.assertEqual(label(z2.imag),None)
        
        del z1
        del z2

    def testIntermediateNaming(self):

        z1 = ucomplex(1,(1,1))
        z2 = ucomplex(1,(1,1))
        y1 = result(z1 + z2)
        self.assertEqual(label(y1),None)
        self.assertEqual(label(y1.real),None)
        self.assertEqual(label(y1.imag),None)
        
        y1 = result(z1 + z2, label='y1')
        self.assertEqual(label(y1),'y1')
        self.assertEqual(label(y1.real),'y1_re')
        self.assertEqual(label(y1.imag),'y1_im')
        
        del z1,z2,y1

#----------------------------------------------------------------------------
class ArithmeticTestsComplex(unittest.TestCase):
    def setUp(self):
        self.x1 = 1+2j
        self.u1 = (.5,0,0,.2)
        self.un1 = ucomplex(self.x1,self.u1)
        
        # Added to extend coverage to _d_components
        self.un1b = ucomplex(self.x1,self.u1,independent=False)
        
        self.x2 = 2+2j
        self.u2 = (1.5,.7)
        self.un2 = ucomplex(self.x2,self.u2)
        
        self.x3 = 3+3j
        self.u3 = (.05,1.2)
        self.un3 = ucomplex(self.x3,self.u3)

        self.x4 = 16
        self.u4 = 4.5
        self.un4 = ureal(self.x4,self.u4)
        
        # Added to extend coverage to _d_components 
        self.un4b = ureal(self.x4,self.u4,independent=False)
 
        # Use these as arguments for intermediate testing
        self.un5 = result( (self.un1 + self.un2) * self.un3 )
        self.u5 = uncertainty(self.un5)

        # Added to extend coverage to _d_components 
        self.un5b = result( (self.un1b + self.un2) * self.un3 )

        self.un6 = result( self.un4 * ureal(-45,4) )
        self.u6 = uncertainty(self.un6)
 
        self.un6b = result( self.un4b * ureal(-45,4) )
        
    def testAddition(self):
        # Cases to consider: uc-uc, uc-numb, numb-uc, uc-ur, ur-uc
        # Need to check u_component in each case.

        # uc-uc
        y = self.un1 + self.un2
        equivalent_complex(value(y),self.x1+self.x2,TOL)
        equivalent_sequence(
            u_component(y,self.un1)
        ,   [ math.sqrt(self.u1[0]),0.,0.,math.sqrt(self.u1[3])]
        ,   TOL
        )
        equivalent_sequence(
            u_component(y,self.un2),
            [self.u2[0],0.,0.,self.u2[1]],
            TOL)
        
        y += self.un3
        equivalent_complex(value(y),self.x1+self.x2+self.x3,TOL)
        equivalent_sequence(
            u_component(y,self.un1)
        ,   [math.sqrt(self.u1[0]),0.,0.,math.sqrt(self.u1[3])]
        ,   TOL
        )
        equivalent_sequence(u_component(y,self.un2),[self.u2[0],0.,0.,self.u2[1]],TOL)
        equivalent_sequence(u_component(y,self.un3),[self.u3[0],0.,0.,self.u3[1]],TOL)

        # Test _d_components
        # uc-uc
        y = self.un1b + self.un2
        equivalent_complex(value(y),self.x1+self.x2,TOL)
        equivalent_sequence(
            u_component(y,self.un1b)
        ,   [ math.sqrt(self.u1[0]),0.,0.,math.sqrt(self.u1[3])]
        ,   TOL
        )
        equivalent_sequence(
            u_component(y,self.un2),
            [self.u2[0],0.,0.,self.u2[1]],
            TOL)
        
        y += self.un3
        equivalent_complex(value(y),self.x1+self.x2+self.x3,TOL)
        equivalent_sequence(
            u_component(y,self.un1b)
        ,   [math.sqrt(self.u1[0]),0.,0.,math.sqrt(self.u1[3])]
        ,   TOL
        )

        # Intermediate u_component test
        y = self.un2 + self.un5        
        equivalent_sequence(
            u_component(y,self.un5)
        ,   [self.u5[0],0,0,self.u5[1]]
        ,   TOL
        )

        # Add to number on left and 0 on left
        num = 5.5
        y = num + self.un2
        equivalent_complex(value(y),num+self.x2,TOL)
        equivalent_sequence(u_component(y,self.un2),[self.u2[0],0.,0.,self.u2[1]],TOL)

        # Intermediate u_component test
        y = num + self.un5        
        equivalent_sequence(
            u_component(y,self.un5)
        ,   [self.u5[0],0,0,self.u5[1]]
        ,   TOL
        )

        num = 0.0
        y = num + self.un2
        self.assertTrue(y is self.un2)
        equivalent_complex(value(y),num+self.x2,TOL)
        equivalent_sequence(u_component(y,self.un2),[self.u2[0],0.,0.,self.u2[1]],TOL)
        

        # Add to number on right and 0 on right
        num = 2.7
        y = self.un1 + num
        equivalent_complex(value(y),self.x1+num,TOL)
        equivalent_sequence(
            u_component(y,self.un1)
        ,   [math.sqrt(self.u1[0]),0.,0.,math.sqrt(self.u1[3])]
        ,   TOL
        )
        
        # Intermediate u_component test
        y = self.un5 + num     
        equivalent_sequence(
            u_component(y,self.un5)
        ,   [self.u5[0],0,0,self.u5[1]]
        ,   TOL
        )

        num = 0.0
        y = self.un1 + num
        self.assertTrue(y is self.un1)
        equivalent_complex(value(y),self.x1+num,TOL)
        equivalent_sequence(
            u_component(y,self.un1)
        ,   [math.sqrt(self.u1[0]),0.,0.,math.sqrt(self.u1[3])]
        ,   TOL
        )
        
        num = 0+2j
        y = self.un1 + num
        self.assertTrue(y.real is not self.un1)
        self.assertTrue(not y.is_elementary)
        equivalent_complex(value(y),self.x1+num,TOL)
        
        # uc-ur and ur-uc
        y = self.un1 + self.un4
        equivalent_complex(value(y),self.x1+self.x4,TOL)
        equivalent_sequence(
            u_component(y,self.un1)
        ,   [ math.sqrt(self.u1[0]),0.,0.,math.sqrt(self.u1[3])]
        ,   TOL
        )
        equivalent_sequence(
            u_component(y,self.un4),
            [self.u4,0.,0.,0.0],
            TOL)

        y = self.un4 + self.un1 
        equivalent_complex(value(y),self.x1+self.x4,TOL)
        equivalent_sequence(
            u_component(y,self.un1)
        ,   [ math.sqrt(self.u1[0]),0.,0.,math.sqrt(self.u1[3])]
        ,   TOL
        )
        equivalent_sequence(
            u_component(y,self.un4),
            [self.u4,0.,0.,0.0],
            TOL)
        
        # Intermediate u_component test
        y = self.un6 + self.un5        
        equivalent_sequence(
            u_component(y,self.un5)
        ,   [self.u5[0],0,0,self.u5[1]]
        ,   TOL
        )
        equivalent_sequence(
            u_component(y,self.un6)
        ,   [self.u6,0,0,0]
        ,   TOL
        )
        y = self.un5 + self.un6         
        equivalent_sequence(
            u_component(y,self.un5)
        ,   [self.u5[0],0,0,self.u5[1]]
        ,   TOL
        )
        equivalent_sequence(
            u_component(y,self.un6)
        ,   [self.u6,0,0,0]
        ,   TOL
        )

        # Extra tests with _d_components
        # uc-ur and ur-uc
        y = self.un1 + self.un4
        y = self.un4b + self.un1 
        equivalent_complex(value(y),self.x1+self.x4,TOL)
        equivalent_sequence(
            u_component(y,self.un1)
        ,   [ math.sqrt(self.u1[0]),0.,0.,math.sqrt(self.u1[3])]
        ,   TOL
        )
        equivalent_sequence(
            u_component(y,self.un4b),
            [self.u4,0.,0.,0.0],
            TOL)
        
        # Intermediate u_component test
        y = self.un6b + self.un5        
        equivalent_sequence(
            u_component(y,self.un5)
        ,   [self.u5[0],0,0,self.u5[1]]
        ,   TOL
        )
        equivalent_sequence(
            u_component(y,self.un6b)
        ,   [self.u6,0,0,0]
        ,   TOL
        )
        y = self.un5 + self.un6b         
        equivalent_sequence(
            u_component(y,self.un5)
        ,   [self.u5[0],0,0,self.u5[1]]
        ,   TOL
        )
        equivalent_sequence(
            u_component(y,self.un6b)
        ,   [self.u6,0,0,0]
        ,   TOL
        )        
    def testSubtraction(self):
        # Cases to consider: uc-uc, uc-numb, numb-uc, uc-ur, ur-uc
        # Need to check u_component in each case.

        y = self.un1 - self.un2
        equivalent_complex(value(y),self.x1-self.x2,TOL)
        equivalent_sequence(
            u_component(y,self.un1)
        ,   [math.sqrt(self.u1[0]),0.,0.,math.sqrt(self.u1[3])]
        ,   TOL
        )
        equivalent_sequence(u_component(y,self.un2),[-self.u2[0],0.,0.,-self.u2[1]],TOL)
        
        y -= self.un3
        equivalent_complex(value(y),self.x1-self.x2-self.x3,TOL)
        equivalent_sequence(
            u_component(y,self.un1)
        ,   [math.sqrt(self.u1[0]),0.,0.,math.sqrt(self.u1[3])]
        ,   TOL
        )
        equivalent_sequence(u_component(y,self.un2),[-self.u2[0],0.,0.,-self.u2[1]],TOL)
        equivalent_sequence(u_component(y,self.un3),[-self.u3[0],0.,0.,-self.u3[1]],TOL)

        # Test _d_components 
        y = self.un1b - self.un2
        equivalent_complex(value(y),self.x1-self.x2,TOL)
        equivalent_sequence(
            u_component(y,self.un1b)
        ,   [math.sqrt(self.u1[0]),0.,0.,math.sqrt(self.u1[3])]
        ,   TOL
        )
        equivalent_sequence(u_component(y,self.un2),[-self.u2[0],0.,0.,-self.u2[1]],TOL)
        
        y -= self.un3
        equivalent_complex(value(y),self.x1-self.x2-self.x3,TOL)
        equivalent_sequence(
            u_component(y,self.un1b)
        ,   [math.sqrt(self.u1[0]),0.,0.,math.sqrt(self.u1[3])]
        ,   TOL
        )
        
        # Intermediate u_component test
        y = self.un2 - self.un5        
        equivalent_sequence(
            u_component(y,self.un5)
        ,   [-self.u5[0],0,0,-self.u5[1]]
        ,   TOL
        )

        # Sub from number on left
        num = 3.5
        y = num - self.un2
        equivalent_complex(value(y),num-self.x2,TOL)
        equivalent_sequence(
            u_component(y,self.un2)
        ,   [-self.u2[0],0.,0.,-self.u2[1]]
        ,   TOL
        )
        
        # Sub from number number on right
        num = 3.5
        y = self.un1 - num
        equivalent_complex(value(y),self.x1-num,TOL)
        equivalent_sequence(
            u_component(y,self.un1)
        ,   [math.sqrt(self.u1[0]),0.,0.,math.sqrt(self.u1[3])]
        ,   TOL
        )
        
        # Intermediate u_component test
        y = self.un5 - num     
        equivalent_sequence(
            u_component(y,self.un5)
        ,   [self.u5[0],0,0,self.u5[1]]
        ,   TOL
        )

        # Sub from 0 on right
        y = self.un1 - 0.0+0.0j
        self.assertTrue(y is self.un1)
        equivalent_complex(value(y),self.x1,TOL)
        equivalent_sequence(
            u_component(y,self.un1)
        ,   [math.sqrt(self.u1[0]),0.,0.,math.sqrt(self.u1[3])]
        ,   TOL
        )

        y = self.un1 - 0.0
        self.assertTrue(y is self.un1)
        equivalent_complex(value(y),self.x1,TOL)
        equivalent_sequence(
            u_component(y,self.un1)
        ,   [math.sqrt(self.u1[0]),0.,0.,math.sqrt(self.u1[3])]
        ,   TOL
        )

        num = 0+2j
        y = self.un1 - num
        self.assertTrue(y.real is not self.un1)
        self.assertTrue(not y.is_elementary)
        equivalent_complex(value(y),self.x1-num,TOL)

        # Sub from 0 on left
        y = 0+0j - self.un1
        self.assertTrue(y is not self.un1)
        equivalent_complex(value(y),-self.x1,TOL)

        num = 0+2j
        y = num - self.un1
        self.assertTrue(y.real is not self.un1)
        self.assertTrue(not y.is_elementary)
        equivalent_complex(value(y),num-self.x1,TOL)

        # uc-ur and ur-uc
        y = self.un1 - self.un4
        equivalent_complex(value(y),self.x1-self.x4,TOL)
        equivalent_sequence(
            u_component(y,self.un1)
        ,   [ math.sqrt(self.u1[0]),0.,0.,math.sqrt(self.u1[3])]
        ,   TOL
        )
        equivalent_sequence(
            u_component(y,self.un4),
            [-self.u4,0.,0.,0.0],
            TOL)

        y = self.un4 - self.un1 
        equivalent_complex(value(y),self.x4-self.x1,TOL)
        equivalent_sequence(
            u_component(y,self.un1)
        ,   [ -math.sqrt(self.u1[0]),0.,0.,-math.sqrt(self.u1[3])]
        ,   TOL
        )
        equivalent_sequence(
            u_component(y,self.un4),
            [self.u4,0.,0.,0.0],
            TOL)

        # Extra tests with _d_component 
        # uc-ur and ur-uc
        y = self.un1 - self.un4b
        equivalent_complex(value(y),self.x1-self.x4,TOL)
        equivalent_sequence(
            u_component(y,self.un1)
        ,   [ math.sqrt(self.u1[0]),0.,0.,math.sqrt(self.u1[3])]
        ,   TOL
        )
        equivalent_sequence(
            u_component(y,self.un4b),
            [-self.u4,0.,0.,0.0],
            TOL)

        y = self.un4b - self.un1 
        equivalent_complex(value(y),self.x4-self.x1,TOL)
        equivalent_sequence(
            u_component(y,self.un1)
        ,   [ -math.sqrt(self.u1[0]),0.,0.,-math.sqrt(self.u1[3])]
        ,   TOL
        )
        equivalent_sequence(
            u_component(y,self.un4b),
            [self.u4,0.,0.,0.0],
            TOL)   
            
        # Intermediate u_component test
        y = self.un6 - self.un5        
        equivalent_sequence(
            u_component(y,self.un5)
        ,   [-self.u5[0],0,0,-self.u5[1]]
        ,   TOL
        )
        equivalent_sequence(
            u_component(y,self.un6)
        ,   [self.u6,0,0,0]
        ,   TOL
        )
        y = self.un5 - self.un6         
        equivalent_sequence(
            u_component(y,self.un5)
        ,   [self.u5[0],0,0,self.u5[1]]
        ,   TOL
        )
        equivalent_sequence(
            u_component(y,self.un6)
        ,   [-self.u6,0,0,0]
        ,   TOL
        )

    def testNegation(self):
        x = self.un1
        w = self.un2
        z = result(x + w)
        y = -(z)
        equivalent_complex(value(y),-value(value(self.un1)+value(self.un2)),TOL)
        u = uncertainty(x)
        equivalent_sequence(
            u_component(y,x),
            [-u[0],0,0,-u[1]],
            TOL)
        u = uncertainty(w)
        equivalent_sequence(
            u_component(y,w),
            [-u[0],0,0,-u[1]],
            TOL)

        # intermediate
        u = uncertainty(z)
        equivalent_sequence(
            u_component(y,z),
            [-u[0],0,0,-u[1]],
            TOL)

    def testMultiplication(self):
        # Cases to consider: uc-uc, uc-numb, numb-uc, uc-ur, ur-uc
        # Need to check u_component in each case.

        y = self.un1 * self.un2
        equivalent_complex(value(y),self.x1*self.x2,TOL)
        u1,r = to_std_uncertainty(self.u1)
        equivalent_sequence(
            u_component(y,self.un1),
            array_to_sequence( numpy.matmul(number_to_matrix(self.x2), u1)),
            TOL)
        u2,r = to_std_uncertainty(self.u2)
        equivalent_sequence(
            u_component(y,self.un2),
            array_to_sequence( numpy.matmul(number_to_matrix(self.x1), u2)),
            TOL)
        
        y *= self.un3
        equivalent_complex(value(y),self.x1*self.x2*self.x3,TOL)
        u3,r = to_std_uncertainty(self.u3)
        equivalent_sequence(
            u_component(y,self.un1),
            array_to_sequence( numpy.matmul(number_to_matrix(self.x2*self.x3), u1)),
            TOL)
        equivalent_sequence(
            u_component(y,self.un2),
            array_to_sequence( numpy.matmul(number_to_matrix(self.x1*self.x3), u2)),
            TOL)
        equivalent_sequence(
            u_component(y,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(self.x2*self.x1), u3)),
            TOL)

        # Test _d_components
        y = self.un1b * self.un2
        equivalent_complex(value(y),self.x1*self.x2,TOL)
        u1,r = to_std_uncertainty(self.u1)
        equivalent_sequence(
            u_component(y,self.un1b),
            array_to_sequence( numpy.matmul(number_to_matrix(self.x2), u1)),
            TOL)
        u2,r = to_std_uncertainty(self.u2)
        equivalent_sequence(
            u_component(y,self.un2),
            array_to_sequence( numpy.matmul(number_to_matrix(self.x1), u2)),
            TOL)
        
        y *= self.un3
        equivalent_complex(value(y),self.x1*self.x2*self.x3,TOL)
        u3,r = to_std_uncertainty(self.u3)
        equivalent_sequence(
            u_component(y,self.un1b),
            array_to_sequence( numpy.matmul(number_to_matrix(self.x2*self.x3), u1)),
            TOL)
        equivalent_sequence(
            u_component(y,self.un2),
            array_to_sequence( numpy.matmul(number_to_matrix(self.x1*self.x3), u2)),
            TOL)
        equivalent_sequence(
            u_component(y,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(self.x2*self.x1), u3)),
            TOL)
            
        # Intermediate u_component test
        y = self.un2 * self.un5
        u5,r = to_std_uncertainty(self.u5)
        equivalent_sequence(
            u_component(y,self.un5)
        ,   array_to_sequence( numpy.matmul(number_to_matrix(self.x2), u5))
        ,   TOL
        )

        # Mul to number on left, mul to 1 on left  
        num = -3.2-5j
        y = num * self.un2
        equivalent_complex(value(y),num*self.x2,TOL)
        u2,r = to_std_uncertainty(self.u2)
        equivalent_sequence(
            u_component(y,self.un2),
            array_to_sequence( numpy.matmul(number_to_matrix(num), u2)),
            TOL)

        # Intermediate u_component test
        y = num * self.un5
        u5,r = to_std_uncertainty(self.u5)
        equivalent_sequence(
            u_component(y,self.un5)
        ,   array_to_sequence( numpy.matmul(number_to_matrix(num), u5))
        ,   TOL
        )

        num = 1.0
        y = num * self.un2
        self.assertTrue(y is self.un2)
        equivalent_complex(value(y),num*self.x2,TOL)
        u2,r = to_std_uncertainty(self.u2)
        equivalent_sequence(
            u_component(y,self.un2),
            array_to_sequence( numpy.matmul(number_to_matrix(num), u2)),
            TOL)

        # Mul to number on right, mul and 1 on right
        num = -3.2+8j
        y = self.un1 * num
        equivalent_complex(value(y),self.x1*num,TOL)
        u1,r = to_std_uncertainty(self.u1)
        equivalent_sequence(
            u_component(y,self.un1),
            array_to_sequence( numpy.matmul(number_to_matrix(num), u1)),
            TOL)
        
        # Intermediate u_component test
        y = self.un5 * num
        u5,r = to_std_uncertainty(self.u5)
        equivalent_sequence(
            u_component(y,self.un5)
        ,   array_to_sequence( numpy.matmul(number_to_matrix(num), u5))
        ,   TOL
        )

        num = 1.0
        y = self.un1 * num
        self.assertTrue(y is self.un1)
        equivalent_complex(value(y),self.x1*num,TOL)
        u1,r = to_std_uncertainty(self.u1)
        equivalent_sequence(
            u_component(y,self.un1),
            array_to_sequence( numpy.matmul(number_to_matrix(num), u1)),
            TOL)

        # uc-ur and ur-uc
        y = self.un1 *self.un4
        equivalent_complex(value(y),self.x1*self.x4,TOL)
        u1,r = to_std_uncertainty(self.u1)
        equivalent_sequence(
            u_component(y,self.un1),
            array_to_sequence( numpy.matmul(number_to_matrix(self.x4), u1))
        ,   TOL
        )
        u4 = numpy.asarray( [ [self.u4,0], [0,0] ])
        equivalent_sequence(
            u_component(y,self.un4),
            array_to_sequence( numpy.matmul(number_to_matrix(self.x1), u4)),
            TOL)

        y = self.un4 * self.un1 
        equivalent_complex(value(y),self.x4*self.x1,TOL)
        u1,r = to_std_uncertainty(self.u1)
        equivalent_sequence(
            u_component(y,self.un1),
            array_to_sequence( numpy.matmul(number_to_matrix(self.x4), u1)),
            TOL
        )
        u4 = numpy.asarray( [ [self.u4,0], [0,0] ])
        equivalent_sequence(
            u_component(y,self.un4),
            array_to_sequence( numpy.matmul(number_to_matrix(self.x1), u4)),
            TOL
        )
        
        # Additional test with _d_components
        # uc-ur and ur-uc
        y = self.un1 *self.un4b
        equivalent_complex(value(y),self.x1*self.x4,TOL)
        u1,r = to_std_uncertainty(self.u1)
        equivalent_sequence(
            u_component(y,self.un1),
            array_to_sequence( numpy.matmul(number_to_matrix(self.x4), u1))
        ,   TOL
        )
        u4 = numpy.asarray( [ [self.u4,0], [0,0] ])
        equivalent_sequence(
            u_component(y,self.un4b),
            array_to_sequence( numpy.matmul(number_to_matrix(self.x1), u4)),
            TOL)

        y = self.un4b * self.un1 
        equivalent_complex(value(y),self.x4*self.x1,TOL)
        u1,r = to_std_uncertainty(self.u1)
        equivalent_sequence(
            u_component(y,self.un1),
            array_to_sequence( numpy.matmul(number_to_matrix(self.x4), u1)),
            TOL
        )
        u4 = numpy.asarray( [ [self.u4,0], [0,0] ])
        equivalent_sequence(
            u_component(y,self.un4b),
            array_to_sequence( numpy.matmul(number_to_matrix(self.x1), u4)),
            TOL
        )
        
        # Intermediate u_component test
        y = self.un6 * self.un5
        u6 = self.u6
        u5,r = to_std_uncertainty(self.u5)    
        equivalent_sequence(
            u_component(y,self.un5)
        ,   array_to_sequence( numpy.matmul(number_to_matrix( value(self.un6) ), u5))
        ,   TOL
        )
        equivalent_sequence(
            u_component(y,self.un6)
        ,   array_to_sequence( numpy.matmul(number_to_matrix( value(self.un5) ), numpy.asarray( [[u6,0],[0,0]] )))
        ,   TOL
        )
        y = self.un5 * self.un6         
        equivalent_sequence(
            u_component(y,self.un5)
        ,   array_to_sequence( numpy.matmul(number_to_matrix( value(self.un6) ), u5))
        ,   TOL
        )
        equivalent_sequence(
            u_component(y,self.un6)
        ,   array_to_sequence( numpy.matmul(number_to_matrix( value(self.un5) ), numpy.asarray( [[u6,0],[0,0]] )))
        ,   TOL
        )

    def testDivision(self):
        # Cases to consider: uc-uc, uc-numb, numb-uc, uc-ur, ur-uc
        # Need to check u_component in each case.

        y = self.un1 / self.un2
        equivalent_complex(value(y),self.x1/self.x2,TOL)
        u1,r = to_std_uncertainty(self.u1)
        u2,r = to_std_uncertainty(self.u2)
        equivalent_sequence(
            u_component(y,self.un1) 
        ,   array_to_sequence( numpy.matmul(number_to_matrix( value(y) / self.x1 ), u1 ))
        ,   TOL
        )
        equivalent_sequence(
            u_component(y,self.un2) 
        ,   array_to_sequence( numpy.matmul(number_to_matrix( -value(y) / self.x2 ), u2 ))
        ,   TOL
        )

        y /= self.un3        
        u3,r = to_std_uncertainty(self.u3)
        equivalent_complex(value(y),self.x1/self.x2/self.x3,TOL)
        equivalent_sequence(
            u_component(y,self.un3) 
        ,   array_to_sequence( numpy.matmul(number_to_matrix( -value(y) / self.x3 ), u3 ))
        ,   TOL
        )

        self.assertRaises(ZeroDivisionError,div,y,0)

        # Test for _d_component 
        y = self.un1b / self.un2
        equivalent_complex(value(y),self.x1/self.x2,TOL)
        u1,r = to_std_uncertainty(self.u1)
        u2,r = to_std_uncertainty(self.u2)
        equivalent_sequence(
            u_component(y,self.un1b) 
        ,   array_to_sequence( numpy.matmul(number_to_matrix( value(y) / self.x1 ), u1 ))
        ,   TOL
        )
        equivalent_sequence(
            u_component(y,self.un2) 
        ,   array_to_sequence( numpy.matmul(number_to_matrix( -value(y) / self.x2 ), u2 ))
        ,   TOL
        )

        y /= self.un3        
        u3,r = to_std_uncertainty(self.u3)
        equivalent_complex(value(y),self.x1/self.x2/self.x3,TOL)
        equivalent_sequence(
            u_component(y,self.un3) 
        ,   array_to_sequence( numpy.matmul(number_to_matrix( -value(y) / self.x3 ), u3 ))
        ,   TOL
        )
        
        # Intermediate u_component test
        y = self.un2 / self.un5
        u5,r = to_std_uncertainty(self.u5)
        equivalent_sequence(
            u_component(y,self.un5)
        ,   array_to_sequence( numpy.matmul(number_to_matrix( -value(y)/value(self.un5) ), u5))
        ,   TOL
        )

        y = self.un5 / self.un2
        u5,r = to_std_uncertainty(self.u5)
        equivalent_sequence(
            u_component(y,self.un5)
        ,   array_to_sequence( numpy.matmul(number_to_matrix( value(y)/value(self.un5) ), u5))
        ,   TOL
        )

        # Div from number on left
        num = 3 - 5j
        y = num / self.un2
        equivalent_complex(value(y),num/self.x2,TOL)
        u2,r = to_std_uncertainty(self.u2)
        equivalent_sequence(
            u_component(y,self.un2) 
        ,   array_to_sequence( numpy.matmul(number_to_matrix( -value(y) / self.x2 ), u2 ))
        ,   TOL
        )

        y = num / self.un5
        u5,r = to_std_uncertainty(self.u5)
        equivalent_sequence(
            u_component(y,self.un5)
        ,   array_to_sequence( numpy.matmul(number_to_matrix(-value(y)/value(self.un5)), u5))
        ,   TOL
        )

        # Div from number on right and 1.0 on right
        num = 2.1 + 7.3j
        y = self.un1 / num
        equivalent_complex(value(y),self.x1/num,TOL)
        u1,r = to_std_uncertainty(self.u1)
        equivalent_sequence(
            u_component(y,self.un1) 
        ,   array_to_sequence( numpy.matmul(number_to_matrix( value(y) / self.x1 ), u1 ))
        ,   TOL
        )

        y = self.un1 / 1.0
        self.assertTrue(y is self.un1)
        equivalent_complex(value(y),self.x1,TOL)
        equivalent_sequence(
            u_component(y,self.un1) 
        ,   array_to_sequence( numpy.matmul(number_to_matrix( 1.0 ), u1 ))
        ,   TOL
        )

        y = self.un5 / num
        u5,r = to_std_uncertainty(self.u5)
        equivalent_sequence(
            u_component(y,self.un5)
        ,   array_to_sequence( numpy.matmul(number_to_matrix(value(y)/value(self.un5)), u5))
        ,   TOL
        )

        # uc-ur and ur-uc
        y = self.un1 / self.un4
        equivalent_complex(value(y),self.x1/self.x4,TOL)
        u1,r = to_std_uncertainty(self.u1)
        equivalent_sequence(
            u_component(y,self.un1),
            array_to_sequence( numpy.matmul(number_to_matrix( value(y) / self.x1 ), u1))
        ,   TOL
        )
        u4 = numpy.asarray( [ [self.u4,0], [0,0] ])
        equivalent_sequence(
            u_component(y,self.un4),
            array_to_sequence( numpy.matmul(number_to_matrix(-value(y) / self.x4), u4)),
            TOL)

        y = self.un4 / self.un1 
        equivalent_complex(value(y),self.x4/self.x1,TOL)
        u1,r = to_std_uncertainty(self.u1)
        equivalent_sequence(
            u_component(y,self.un1),
            array_to_sequence( numpy.matmul(number_to_matrix( -value(y) / self.x1 ), u1)),
            TOL
        )
        u4 = numpy.asarray( [ [self.u4,0], [0,0] ])
        equivalent_sequence(
            u_component(y,self.un4),
            array_to_sequence( numpy.matmul(number_to_matrix( value(y) / self.x4), u4)),
            TOL
        )

        # Additional tests with _d_components 
        # uc-ur and ur-uc
        y = self.un1 / self.un4b
        equivalent_complex(value(y),self.x1/self.x4,TOL)
        u1,r = to_std_uncertainty(self.u1)
        equivalent_sequence(
            u_component(y,self.un1),
            array_to_sequence( numpy.matmul(number_to_matrix( value(y) / self.x1 ), u1))
        ,   TOL
        )
        u4 = numpy.asarray( [ [self.u4,0], [0,0] ])
        equivalent_sequence(
            u_component(y,self.un4b),
            array_to_sequence( numpy.matmul(number_to_matrix(-value(y) / self.x4), u4)),
            TOL)

        y = self.un4b / self.un1 
        equivalent_complex(value(y),self.x4/self.x1,TOL)
        u1,r = to_std_uncertainty(self.u1)
        equivalent_sequence(
            u_component(y,self.un1),
            array_to_sequence( numpy.matmul(number_to_matrix( -value(y) / self.x1 ), u1)),
            TOL
        )
        u4 = numpy.asarray( [ [self.u4,0], [0,0] ])
        equivalent_sequence(
            u_component(y,self.un4b),
            array_to_sequence( numpy.matmul(number_to_matrix( value(y) / self.x4), u4)),
            TOL
        )
        # Intermediate u_component test
        y = self.un6 / self.un5
        u6 = self.u6
        u5,r = to_std_uncertainty(self.u5)    
        equivalent_sequence(
            u_component(y,self.un5)
        ,   array_to_sequence( numpy.matmul(number_to_matrix( -value(y)/value(self.un5) ), u5))
        ,   TOL
        )
        equivalent_sequence(
            u_component(y,self.un6)
        ,   array_to_sequence( numpy.matmul(number_to_matrix( value(y)/value(self.un6) ), numpy.asarray( [[u6,0],[0,0]])))
        ,   TOL
        )
        y = self.un5 / self.un6         
        equivalent_sequence(
            u_component(y,self.un5)
        ,   array_to_sequence( numpy.matmul(number_to_matrix( value(y)/value(self.un5) ), u5))
        ,   TOL
        )
        equivalent_sequence(
            u_component(y,self.un6)
        ,   array_to_sequence( numpy.matmul(number_to_matrix( -value(y)/value(self.un6) ), numpy.asarray( [[u6,0],[0,0]])))
        ,   TOL
        )

        # Zero-valued numerator
        # Cases to consider: uc-uc, uc-numb, numb-uc, uc-ur, ur-uc
        u = 0.1
        n = ucomplex(0,u)
        d = ucomplex(1,1)
        q = n/d
        dq_dn = 1.0/value(d)
        dq_dd = 0.0
        self.assertTrue(
            equivalent_complex(q.x,0.0)
        )
        equivalent_sequence(
            u_component(q,n)
        ,   [dq_dn*u,-dq_dd*u,dq_dd*u,dq_dn*u]
        ,   TOL
        )

        d = 1
        q = n/d
        dq_dn = 1.0/value(d)
        dq_dd = 0.0
        self.assertTrue(
            equivalent_complex(q.x,0.0)
        )
        equivalent_sequence(
            u_component(q,n)
        ,   [dq_dn*u,-dq_dd*u,dq_dd*u,dq_dn*u]
        ,   TOL
        )

        u = 0.0
        n = 0
        d = ucomplex(1,1)
        q = n/d
        dq_dn = 1.0/value(d)
        dq_dd = 0.0
        self.assertTrue(
            equivalent_complex(q.x,0.0)
        )
        equivalent_sequence(
            u_component(q,n)
        ,   [dq_dn*u,-dq_dd*u,dq_dd*u,dq_dn*u]
        ,   TOL
        )

        z = ucomplex(0,1)
        x = ureal(1,1)
        q = z/x
        dq_dz = 1.0/value(x)
        dq_dx = -value(q)/value(x)
        self.assertTrue( equivalent_complex(value(q),0.0) )
        self.assertTrue(
            equivalent_sequence(u_component(q,z), [dq_dz.real,-dq_dz.imag,dq_dz.imag,dq_dz.real] ) 
        )
 
        z = ucomplex(1,1)
        x = ureal(0,1)
        q = x/z
        dq_dx = 1.0/value(z)
        self.assertTrue( equivalent_complex(value(q),0.0) )
        self.assertTrue(
            equivalent_sequence(u_component(q,x), [dq_dx.real,-dq_dx.imag,0,0] ) 
        )

        # c-uc
        y = self.x1 / self.un4
        equivalent_complex(value(y),self.x1/self.x4,TOL)
        equivalent_sequence(
            u_component(y,self.x1),
            (0.,0.,0.,0.)
        ,   TOL
        )
        u4 = numpy.asarray( [ [self.u4,0], [0,0] ])
        equivalent_sequence(
            u_component(y,self.un4),
            array_to_sequence( numpy.matmul(number_to_matrix(-value(y) / self.x4), u4)),
            TOL)

#-----------------------------------------------------
class GuideExampleH2Complex(unittest.TestCase):

    def setUp(self):

        self.v = ucomplex(4.999,(0.0032,0.0),independent=False)
        self.i = ucomplex(0.019661,(0.0000095,0.0),independent=False)
        self.phi = ucomplex(1.04446,(0.00075,0.0),independent=False)

        set_correlation(-0.36,self.v.real,self.i.real)
        set_correlation(0.86,self.v.real,self.phi.real)
        set_correlation(-0.65,self.i.real,self.phi.real)

    def test(self):
        _J_ = 1j
        TOL = 1E-10
        z = self.v * exp( _J_ * self.phi )/ self.i
        
        equivalent( uncertainty(z)[0],0.0699787279884,TOL)
        equivalent( uncertainty(z)[1],0.295716826846,TOL)
        equivalent( get_correlation(z),-0.591484610819,TOL)
        equivalent( z.r,-0.591484610819,TOL)

        # The willink_hall function provides an alternative
        # calculation of the covariance, so just check that
        # it works.
        cv = willink_hall(z)[0]
        
        u_r = math.sqrt(cv[0])
        u_i = math.sqrt(cv[3])
        r = cv[1]/(u_r*u_i)

        equivalent( u_r,0.0699787279884,TOL)
        equivalent( u_i,0.295716826846,TOL)
        equivalent( r,-0.591484610819,TOL)

#-----------------------------------------------------
class ComplexFunctionTest(unittest.TestCase):

    def setUp(self):
        x1 = 1+2j
        u1 = (.5,0,0,.2)
        
        x2 = 2+2j
        u2 = (1.5,.7)
        
        # An intermediate un-complex for u_component testing
        self.un3 = result( ucomplex(x1,u1) * ucomplex(x2,u2) )

        # An intermediate un-real for u_component testing
        self.un4 = result( ureal(3.2,.4) / ureal(2.2,0.8) )
        
    def test(self):
        # self.python_atrigh_change()
        self.trig()
        self.trig_h()
        self.a_trig()
        self.a_trig_h()
        self.imag_stuff()
        self.misc()
        self.power()
       
    def trig(self):
        v =  -1 + 2j
        mat = numpy.asarray([[0.87,0],[0,0.95] ])
        x = ucomplex(v,( (mat[0,0], mat[1,1]) ) )
        
        s = sin(x)
        c = cos(x)
        t = tan(x)

        # sin -----------------------------------------        
        equivalent_complex(value(s),cmath.sin(v),TOL)
        equivalent_sequence(
            u_component(x,x),
            array_to_sequence( mat ),
            TOL)
        equivalent_sequence(
            u_component(s,x),
            array_to_sequence( numpy.matmul(number_to_matrix( cmath.cos(v) ), mat )),
            TOL)
        self.assertTrue( math.isinf( dof(s) ) )

        # intermediate
        v3 = value(self.un3)
        s3 = sin(self.un3)
        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        equivalent_sequence(
            u_component(s3,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(cmath.cos(v3)), mat3)),
            TOL)

         # cos -----------------------------------------        
        equivalent_complex(value(c),cmath.cos(v),TOL)
        equivalent_sequence(
            u_component(c,x),
            array_to_sequence( numpy.matmul(number_to_matrix(-cmath.sin(v)), mat )),
            TOL)
        self.assertTrue( math.isinf( dof(c) ) )
        
        # intermediate
        v3 = value(self.un3)
        s3 = cos(self.un3)
        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        equivalent_sequence(
            u_component(s3,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(-cmath.sin(v3)), mat3)),
            TOL)

         # tan -----------------------------------------        
        equivalent_complex(value(t),cmath.tan(v),TOL)
        equivalent_sequence(
            u_component(t,x),
            array_to_sequence( numpy.matmul(number_to_matrix(1.0/( cmath.cos(v))**2), mat )),
            TOL)
        self.assertTrue( math.isinf( dof(c) ) )
        
        # intermediate
        v3 = value(self.un3)
        s3 = tan(self.un3)
        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        equivalent_sequence(
            u_component(s3,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(1.0/(cmath.cos(v3))**2), mat3)),
            TOL)
 
    def trig_h(self):
        v = -2 + 0.6j
        mat,r = to_std_uncertainty([1.87,0.,0.,0.55])
        x = ucomplex(v, (mat[0,0],mat[1,1]) )
        
        s = sinh(x)
        c = cosh(x)
        t = tanh(x)
        
        # sinh -----------------------------------------------
        equivalent_complex(value(s),cmath.sinh(v),TOL)
        equivalent_sequence(
            u_component(s,x),
            array_to_sequence( numpy.matmul(number_to_matrix(cmath.cosh(v)), mat)),
            TOL)
        self.assertTrue( math.isinf( dof(s) ) )
        
        # intermediate
        v3 = value(self.un3)
        s3 = sinh(self.un3)
        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        equivalent_sequence(
            u_component(s3,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(cmath.cosh(v3)), mat3)),
            TOL)
        
        
        # cosh -----------------------------------------------      
        equivalent_complex(value(c),cmath.cosh(v),TOL)
        equivalent_sequence(
            u_component(c,x),
            array_to_sequence( numpy.matmul(number_to_matrix(cmath.sinh(v)), mat)),
            TOL)
        self.assertTrue(math.isinf( dof(c) ))

        # intermediate
        v3 = value(self.un3)
        s3 = cosh(self.un3)
        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        equivalent_sequence(
            u_component(s3,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(cmath.sinh(v3)), mat3)),
            TOL)
        
        # tanh -----------------------------------------------      
        equivalent_complex(value(t),cmath.tanh(v),TOL)
        equivalent_sequence(
            u_component(t,x),
            array_to_sequence( numpy.matmul(number_to_matrix(1.0/(cmath.cosh(v))**2), mat )),
            TOL)
        self.assertTrue( math.isinf( dof(t) ) )

        # intermediate
        v3 = value(self.un3)
        s3 = tanh(self.un3)
        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        equivalent_sequence(
            u_component(s3,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(1.0/(cmath.cosh(v3))**2), mat3)),
            TOL)

    def arcos(self,z):
        return -1j * cmath.log(z + 1j*cmath.sqrt(1-z*z))

    def arsin(self,z):
        return -1j * cmath.log(1j*z + cmath.sqrt(1 - z*z))
    
    def artan(self,z):
        return 1j * cmath.log((1j+z)/(1j-z)) / 2

    def a_trig(self):
        v = 0.32 + 1.1j
        mat = numpy.asarray([ [1.17,0],[0,0.35]])
        x = ucomplex(v,(mat[0,0], mat[1,1]) )
        
        s = asin(x)
        c = acos(x)
        t = atan(x)
        
        # asin -----------------------------------------------      
        equivalent_complex(value(s),self.arsin(v),TOL)
        den = 1 - v**2
        equivalent_sequence(
            u_component(s,x),
            array_to_sequence( numpy.matmul(number_to_matrix(1/cmath.sqrt(den)), mat )),
            TOL)
        self.assertTrue( math.isinf( dof(s) ) )

        # intermediate
        v3 = value(self.un3)
        s3 = asin(self.un3)
        den = 1 - v3**2
        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        equivalent_sequence(
            u_component(s3,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(1/cmath.sqrt(den)), mat3)),
            TOL)

        # acos -----------------------------------------------      
        equivalent_complex(value(c),self.arcos(v),TOL)
        den = 1 - v**2
        equivalent_sequence(
            u_component(c,x),
            array_to_sequence( numpy.matmul(number_to_matrix(-1/cmath.sqrt(den)), mat )),
            TOL)
        self.assertTrue( math.isinf( dof(c) ) )

        # intermediate
        v3 = value(self.un3)
        s3 = acos(self.un3)
        den = 1 - v3**2
        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        equivalent_sequence(
            u_component(s3,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(-1/cmath.sqrt(den)), mat3)),
            TOL)

        # atan -----------------------------------------------      
        equivalent_complex(value(t),self.artan(v),TOL)
        den = 1 + v**2
        equivalent_sequence(
            u_component(t,x),
            array_to_sequence( numpy.matmul(number_to_matrix(1.0/den), mat )),
            TOL)
        self.assertTrue( math.isinf( dof(t) ) )

        # intermediate
        v3 = value(self.un3)
        s3 = atan(self.un3)
        den = 1 + v3**2
        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        equivalent_sequence(
            u_component(s3,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(1/den), mat3)),
            TOL)

    def arcosh(self,z):
        return cmath.log(z + cmath.sqrt(z-1)*cmath.sqrt(z+1))

    def arsinh(self,z):
        return cmath.log(z + cmath.sqrt(z*z + 1.0))
    
    def artanh(self,z):
        return cmath.log((1+z)/(1-z)) / 2

    def python_atrigh_change(self):
        # The cmath implementation of inverse hyperbolic trig functions
        # may evolve in future Python releases, so keep an eye on it here.
        z = -2 + 3j
        assert (self.arsinh(z) != cmath.asinh(z)), "Library function has changed"
        for z in (2+3j,-2+3j,2-3j,-2-3j):
            equivalent_complex(self.arcosh(z),cmath.acosh(z),TOL)
            equivalent_complex(self.artanh(z),cmath.atanh(z),TOL)
                
    def a_trig_h(self):
        v = -2 + 3j
        mat = numpy.asarray( [[1.17,0],[0,0.35]] )
        x = ucomplex(v, ((mat[0,0], mat[1,1]) ) )
        
        s = asinh(x)
        c = acosh(x)
        t = atanh(x)
        
        # asinh -----------------------------------------------      
        equivalent_complex(value(s),self.arsinh(v),TOL)
        den = 1 + v**2
        equivalent_sequence(
            u_component(s,x),
            array_to_sequence( numpy.matmul(number_to_matrix(1.0/cmath.sqrt(den)), mat )),
            TOL)
        self.assertTrue( math.isinf( dof(s) ) )

        # intermediate
        v3 = value(self.un3)
        s3 = asinh(self.un3)
        den = 1 + v3**2
        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        equivalent_sequence(
            u_component(s3,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(1/cmath.sqrt(den)), mat3)),
            TOL)

        # acosh -----------------------------------------------      
        equivalent_complex(value(c),cmath.acosh(v),TOL)
        den = cmath.sqrt((v - 1) * (v + 1))
        equivalent_sequence(
            u_component(c,x),
            array_to_sequence( numpy.matmul(number_to_matrix(1.0/den), mat )),
            TOL)
        self.assertTrue( math.isinf( dof(c) ) )

        # intermediate
        v3 = value(self.un3)
        s3 = acosh(self.un3)
        den = cmath.sqrt((v3 - 1) * (v3 + 1))
        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        equivalent_sequence(
            u_component(s3,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(1/(den)), mat3)),
            TOL)

        # atanh -----------------------------------------------      
        equivalent_complex(value(t),cmath.atanh(v),TOL)
        den = 1 - v**2
        equivalent_sequence(
            u_component(t,x),
            array_to_sequence( numpy.matmul(number_to_matrix(1.0/den), mat )),
            TOL)
        self.assertTrue( math.isinf( dof(t) ) )

        # intermediate
        v3 = value(self.un3)
        s3 = atanh(self.un3)
        den = 1 - v3**2
        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        equivalent_sequence(
            u_component(s3,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(1/den), mat3)),
            TOL)

        
    def imag_stuff(self):
        v = -2+8j
        mat,r = to_std_uncertainty([10.17,0,0,11.35])
        x = ucomplex(v, (mat[0,0], mat[1,1]) )

        c = x.conjugate()
        re = x.real
        im = x.imag
        
        # conjugate -----------------------------------------------      
        equivalent_complex(value(c),v.conjugate(), TOL)
        equivalent_sequence(
            u_component(c,x),
            array_to_sequence( numpy.matmul(numpy.asarray([[1,0],[0,-1]]), mat) ),
            TOL)
        self.assertTrue( math.isinf( dof(c) ) )
        
        # intermediate
        s3 = self.un3.conjugate()
        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        equivalent_sequence(
            u_component(s3,self.un3),
            array_to_sequence( numpy.matmul(numpy.asarray([[1,0],[0,-1]]), mat3) ),
            TOL)

        # real -----------------------------------------------      
        equivalent_complex(value(re),complex(v.real),TOL)
        equivalent_sequence(
            u_component(re,x),
            array_to_sequence( numpy.matmul(numpy.asarray([[1,0],[0,0]]), mat) ),
            TOL)
        self.assertTrue( math.isinf( dof(re) ) )

        # intermediate
        s3 = self.un3.real
        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        equivalent_sequence(
            u_component(s3,self.un3),
            array_to_sequence( numpy.matmul(numpy.asarray([[1,0],[0,0]]), mat3) ),
            TOL)

        # imag -----------------------------------------------      
        equivalent_complex(value(im),complex(v.imag),TOL)
        equivalent_sequence(
            u_component(im,x),
            array_to_sequence( numpy.matmul(numpy.asarray([[0,1],[0,0]]), mat) ),
            TOL)
        self.assertTrue( math.isinf( dof(im) ) )

        # intermediate
        s3 = self.un3.imag
        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        equivalent_sequence(
            u_component(s3,self.un3),
            array_to_sequence( numpy.matmul(numpy.asarray([[0,1],[0,0]]), mat3) ),
            TOL)
        
    def misc(self):
        v = 1.1-0.8j
        mat,r = to_std_uncertainty([4.1,0,0,1.7])
        x = ucomplex(v, (mat[0,0], mat[1,1]) )

        l10 = log10(x)
        p = phase(x)
        
        e = exp(x)
        l = log(x)
        s = sqrt(x)
        n = mag_squared(x)
        a = magnitude(x)
        
        z = ucomplex(0, (mat[0,0], mat[1,1]) )
        self.assertRaises(ZeroDivisionError,magnitude,z)

        # log10 -----------------------------------------------------------        
        equivalent_complex(value(l10),cmath.log10(v),TOL)
        equivalent_sequence(
            u_component(l10,x)
        ,   array_to_sequence( numpy.matmul(number_to_matrix(LOG10_E/v), mat ))
        ,   TOL
        )
        self.assertTrue( math.isinf(dof(l10)) )
        
        # exp --------------------------------------------------------
        equivalent_complex(value(e),cmath.exp(v),TOL)
        equivalent_sequence(
            u_component(e,x),
            array_to_sequence( numpy.matmul(number_to_matrix(cmath.exp(v)), mat)),
            TOL)
        self.assertTrue( math.isinf(dof(e)) )
        
        # intermediate
        v3 = value(self.un3)
        s3 = exp(self.un3)
        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        equivalent_sequence(
            u_component(s3,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(cmath.exp(v3)), mat3)),
            TOL)

        # log -----------------------------------------------------------        
        equivalent_complex(value(l),cmath.log(v),TOL)
        equivalent_sequence(
            u_component(l,x),
            array_to_sequence( numpy.matmul(number_to_matrix(1.0/v), mat )),
            TOL)
        self.assertTrue( math.isinf(dof(l)) )

        # intermediate
        v3 = value(self.un3)
        s3 = log(self.un3)
        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        equivalent_sequence(
            u_component(s3,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(1.0/v3), mat3)),
            TOL)

        # sqrt ------------------------------------------------------------
        equivalent_complex(value(s),cmath.sqrt(v),TOL)
        equivalent_sequence(
            u_component(s,x),
            array_to_sequence( numpy.matmul(number_to_matrix(0.5/cmath.sqrt(v)), mat )),
            TOL)
        self.assertTrue( math.isinf( dof(s) ) )

        # intermediate
        v3 = value(self.un3)
        s3 = sqrt(self.un3)
        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        equivalent_sequence(
            u_component(s3,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(0.5/cmath.sqrt(v3)), mat3)),
            TOL)
           
        # mag_squared ------------------------------------------------------------
        equivalent_complex(value(n),complex(abs(v)**2),TOL)
        equivalent_sequence(
            u_component(n,x),
            array_to_sequence( numpy.matmul(numpy.asarray([ [2*v.real,2*v.imag],[0,0] ]), mat )),
            TOL)
        self.assertTrue( math.isinf( dof(n) ) )

        # intermediate
        v3 = value(self.un3)
        s3 = mag_squared(self.un3)
        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        equivalent_sequence(
            u_component(s3,self.un3),
            array_to_sequence( numpy.matmul(numpy.asarray([ [2*v3.real,2*v3.imag],[0,0] ]), mat3)),
            TOL)

        # when argument is ureal, it returns self*self
        v4 = 2
        u4 = 0.5
        z4 = ureal(v4,u4)
        z4_mag = mag_squared(z4)
        self.assertTrue(
            equivalent(value(z4_mag),v4**2)
        )
        self.assertTrue(
            equivalent(uncertainty(z4_mag),2*v4*u4)
        )
        
        # magnitude ------------------------------------------------------------
        equivalent_complex(value(a),complex( abs(v) ),TOL)
        equivalent_sequence(
            u_component(a,x),
            array_to_sequence( numpy.matmul(numpy.asarray([ [v.real/abs(v),v.imag/abs(v)],[0,0] ]), mat )),
            TOL)
        self.assertTrue( math.isinf( dof(a) ) )

        # intermediate
        v3 = value(self.un3)
        s3 = magnitude(self.un3)
        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        equivalent_sequence(
            u_component(s3,self.un3),
            array_to_sequence( numpy.matmul(numpy.asarray([ [v3.real/abs(v3),v3.imag/abs(v3)],[0,0] ]), mat3)),
            TOL)

    def power(self):
        # Cases to consider: uc-uc, uc-numb, numb-uc, uc-ur, ur-uc
        # Need to check u_component in each case.
        
        TOL = 1E-14

        # uc-uc ---------------------------------------------------        
        vx = 2.1-1.8j
        vy = 1.1-0.5j
        matx,r = to_std_uncertainty([4.1,0,0,1.7])
        maty,r = to_std_uncertainty([0.1,0,0,1.1])
        x = ucomplex(vx, (matx[0,0], matx[1,1]) )
        y = ucomplex(vy, (maty[0,0], maty[1,1]) )

        p = x**y
        vp = vx**vy
        equivalent_complex(value(p),vp,TOL)
        
        dp_dx = vp*vy/vx
        equivalent_sequence(
            u_component(p,x),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dx), matx )),
            TOL)
        dp_dy = cmath.log(vx)*vp
        equivalent_sequence(
            u_component(p,y),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dy), maty )),
            TOL)

        # intermediate
        p = x**self.un3
        v3 = value(self.un3)
        vp = vx**v3

        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        dp_dv3 = cmath.log(vx)* vp        
        equivalent_sequence(
            u_component(p,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dv3), mat3 )),
            TOL)

        p = self.un3**y
        vp = v3**vy
        dp_dv3 = vp*vy/v3
        equivalent_sequence(
            u_component(p,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dv3), mat3 )),
            TOL)

        # numb-uc ---------------------------------------------------        
        # x ** n, where x is a number
        vz = 2.5
        p = vz ** y
        vp = vz ** vy
        equivalent_complex(value(p),vp,TOL)
        dp_dy = cmath.log(vz)*vp
        equivalent_sequence(
            u_component(p,y),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dy), maty )),
            TOL)

        # intermediate
        p = vz**self.un3
        v3 = value(self.un3)
        vp = vz**v3

        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        dp_dv3 = cmath.log(vz) * vp
        
        equivalent_sequence(
            u_component(p,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dv3), mat3 )),
            TOL)
        
        # uc-numb ---------------------------------------------------        
        # x ** n, where n is a number        
        vz = 2.5
        p = x ** vz 
        vp = vx ** vz
        equivalent_complex(value(p),vp,TOL)
        dp_dx = vp*vz/vx
        equivalent_sequence(
            u_component(p,x),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dx), matx )),
            TOL)

        # intermediate
        p = self.un3**vz
        v3 = value(self.un3)
        vp = v3**vz

        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        dp_dv3 = vp*vz/v3
        
        equivalent_sequence(
            u_component(p,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dv3), mat3 )),
            TOL)

        # uc-ur ------------------------------------------------
        vx = 2.1-1.8j
        vy = 1.1
        matx,r = to_std_uncertainty([4.1,0,0,1.7])
        uy = 0.12
        x = ucomplex(vx, (matx[0,0], matx[1,1]) )
        y = ureal(vy, uy )

        p = x**y
        vp = vx**vy
        equivalent_complex(value(p),vp,TOL)
        
        dp_dx = vp*vy/vx
        equivalent_sequence(
            u_component(p,x),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dx), matx )),
            TOL)
        dp_dy = cmath.log(vx)*vp
        equivalent_sequence(
            u_component(p,y),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dy), numpy.asarray( [[uy,0],[0,0]] ) )),
            TOL)

        # intermediate
        p = self.un3**self.un4
        v3 = value(self.un3)
        v4 = value(self.un4)
        u4 = uncertainty(self.un4)
        vp = v3**v4

        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        mat4 = numpy.asarray( [[u4,0],[0,0]] )

        dp_dv4 = cmath.log(v3)* vp        
        equivalent_sequence(
            u_component(p,self.un4),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dv4), mat4 )),
            TOL)
        dp_dv3 = vp*v4/v3
        equivalent_sequence(
            u_component(p,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dv3), mat3 )),
            TOL)
        
        # ur-uc ------------------------------------------------
        vx = -1.8
        vy = 1.1-0.5j
        ux = 4.1
        maty,r = to_std_uncertainty([0.1,0,0,1.1])
        x = ureal(vx,ux )
        y = ucomplex(vy, (maty[0,0], maty[1,1]) )

        p = x**y
        vp = vx**vy
        equivalent_complex(value(p),vp,TOL)
        
        dp_dx = vp*vy/vx
        equivalent_sequence(
            u_component(p,x),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dx), numpy.asarray( [[ux,0],[0,0]] ) )),
            TOL)
        dp_dy = cmath.log(vx)*vp
        equivalent_sequence(
            u_component(p,y),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dy), maty )),
            TOL)

        # intermediate
        p = self.un4**self.un3
        v3 = value(self.un3)
        v4 = value(self.un4)
        u4 = uncertainty(self.un4)
        vp = v4**v3

        mat3,r = to_std_uncertainty( uncertainty(self.un3) )
        mat4 = numpy.asarray( [[u4,0],[0,0]] )

        dp_dv3 = cmath.log(v4)* vp        
        equivalent_sequence(
            u_component(p,self.un3),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dv3), mat3 )),
            TOL)
        dp_dv4 = vp*v3/v4
        equivalent_sequence(
            u_component(p,self.un4),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dv4), mat4 )),
            TOL)
        
        # special cases ------------------------------------------------
        vx = 2.1-1.8j
        vy = 1.1-0.5j
        matx,r = to_std_uncertainty([4.1,0,0,1.7])
        maty,r = to_std_uncertainty([0.1,0,0,1.1])
        x = ucomplex(vx, (matx[0,0], matx[1,1]) )
        y = ucomplex(vy, (maty[0,0], maty[1,1]) )

        # x ** n, where n is 0        
        vz = 0.0
        p = x ** vz 
        vp = vx ** vz
        equivalent_complex(value(p),vp,TOL)
        dp_dx = vp*vz/vx
        equivalent_sequence(
            u_component(p,x),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dx), matx )),
            TOL)

        # x ** n, where n is 1
        vz = 1.0
        p = x ** vz
        self.assertTrue( p is x )
        vp = vx ** vz
        equivalent_complex(value(p),vp,TOL)
        dp_dx = vp*vz/vx
        equivalent_sequence(
            u_component(p,x),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dx), matx )),
            TOL)

        # negative real to fractional power ---------------------------------
        # ur-r 
        vx = -1.8
        vy = 0.5
        ux = 4.1
        
        x = ureal( vx,ux )
        y = vy

        p = x**y
        vp = complex(vx,0.)**vy
        equivalent_complex(value(p),vp,TOL)
        
        dp_dx = vp*vy/vx
        equivalent_sequence(
            u_component(p,x),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dx), numpy.asarray( [[ux,0],[0,0]] ) )),
            TOL)
        dp_dy = cmath.log(vx)*vp
        equivalent_sequence(
            u_component(p,y),
            (0.,0.,0.,0.),
            TOL)

        # r-ur 
        vx = -1.8
        vy = 0.5
        uy = 4.1
        maty = numpy.array( (uy,0,0,0) )
        maty.shape = (2,2)
        
        x = vx
        y = ureal( vy,uy )

        p = x**y
        vp = complex(vx,0.)**vy
        equivalent_complex(value(p),vp,TOL)
        
        dp_dx = vp*vy/vx
        equivalent_sequence(
            u_component(p,x),
            (0.,0.,0.,0.),
            TOL)
            
        dp_dy = cmath.log(vx)*vp
        equivalent_sequence(
            u_component(p,y),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dy), maty )),
            TOL)

        # ur-ur 
        vx = -1.8
        vy = 0.5
        ux = 4.1
        uy = 0.4
        
        matx = numpy.array( (ux,0,0,0) )
        matx.shape = (2,2)
        maty = numpy.array( (uy,0,0,0) )
        maty.shape = (2,2)

        x = ureal(vx, ux )
        y = ureal(vy, uy )

        p = x**y
        vp = complex(vx,0)**vy
        equivalent_complex(value(p),vp,TOL)
        
        dp_dx = vp*vy/vx
        equivalent_sequence(
            u_component(p,x),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dx), matx )),
            TOL)
        dp_dy = cmath.log(vx)*vp
        equivalent_sequence(
            u_component(p,y),
            array_to_sequence( numpy.matmul(number_to_matrix(dp_dy), maty )),
            TOL)


#-----------------------------------------------------
class TestMultipleUNsWithComplexConstants(unittest.TestCase):
    """
    When defining ensembles of real or complex uncertain numbers,
    some elements might in fact have no uncertainty! In which
    case they are defined as constants.
    
    This seems a bit artificial: in effect the user is
    passing an uncertain number as part of an ensemble
    when it is not! However, we may have this situation
    in fitting scenarios (perhaps a parameter comes
    back with no uncertainty due to some odd data).
    So it is provided for. The creation of a constant
    should not harm subsequent calculations.
    
    """
        
    def test_ucomplex(self):
        values = [4.999+0j,0.019661+0j,1.04446j]
        uncert = [(0.0032,0.3),(0,0),(0.2,0.00075)]
        
        v,i,phi = multiple_ucomplex(values,uncert,5)

        self.assertTrue( equivalent_complex(v.x, values[0]) )
        self.assertTrue( equivalent_sequence(v.u, uncert[0]) )
        self.assertTrue(v.df == 5)

        self.assertTrue( equivalent_complex(i.x, values[1]) )
        self.assertTrue( equivalent_sequence(i.u, uncert[1]) )
        self.assertEqual(i.df, inf)

        self.assertTrue( equivalent_complex(phi.x, values[2]) )
        self.assertTrue( equivalent_sequence(phi.u, uncert[2]) )
        self.assertEqual(phi.df, 5)

        # We can set a zero correlation 
        set_correlation([0,0,0,0],v,i)
        set_correlation([0,0,0,0],i,phi)
        set_correlation([0,0,0,0],phi,v)
        
        # and set correlation between pairs in the ensemble
        set_correlation([0.5,0.1,0.2,0.3],v,phi)
        
#-----------------------------------------------------
class TestComplexUncertainRealMath(unittest.TestCase):

    def test_addition(self):
        x = ureal(1,1)
        a = 3
        b = 0.5
        z = complex(a,b)

        y = x + z
        self.assertTrue( isinstance(y,UncertainComplex) )
        
        self.assertTrue( equivalent_complex( complex(1+a,b), value(y), TOL) )
        self.assertTrue( equivalent( 1, uncertainty(y).real, TOL) )
        self.assertTrue( equivalent( 0, uncertainty(y).imag, TOL) )

        # components
        self.assertTrue( equivalent_sequence( u_component(y,x), (1,0,0,0) ) )
        self.assertTrue( equivalent_sequence( u_component(y,z), (0,0,0,0) ) )

        # same thing
        xz = ucomplex(1,(1,0))
        yz = xz + z
        self.assertTrue( equivalent_complex( value(yz), value(y), TOL) )
        self.assertTrue( equivalent( uncertainty(yz).real, uncertainty(y).real, TOL) )
        self.assertTrue( equivalent( uncertainty(yz).imag, uncertainty(y).imag, TOL) )

        # complex on right        
        y = z + x
        self.assertTrue( isinstance(y,UncertainComplex) )

        self.assertTrue( equivalent_complex( complex(1+a,b), value(y), TOL) )
        self.assertTrue( equivalent( 1, uncertainty(y).real, TOL) )
        self.assertTrue( equivalent( 0, uncertainty(y).imag, TOL) )
        
        # components
        self.assertTrue( equivalent_sequence( u_component(y,x), (1,0,0,0) ) )
        self.assertTrue( equivalent_sequence( u_component(y,z), (0,0,0,0) ) )

        # same thing
        yz = z + xz
        self.assertTrue( equivalent_complex( value(yz), value(y), TOL) )
        self.assertTrue( equivalent( uncertainty(yz).real, uncertainty(y).real, TOL) )
        self.assertTrue( equivalent( uncertainty(yz).imag, uncertainty(y).imag, TOL) )

        # intermediate components
        # -- on the right
        a = 3
        b = 0.5
        z = complex(a,b)
        x1v = 3.1
        x2v = 4.1
        x1 = result(ureal(x1v,1) + ureal(x2v,1) )
        y = x1 + z

        ux1 = uncertainty(x1)
        self.assertTrue( equivalent_sequence( u_component(y,x1), (ux1,0,0,0) ) )

        # -- on the left
        y = z + x1
        self.assertTrue( equivalent_sequence( u_component(y,x1), (ux1,0,0,0) ) )
      

    def test_subtraction(self):
        x = ureal(1,1.0)
        a = 3
        b = 0.5
        z = complex(a,b)

        y = x - z
        self.assertTrue( isinstance(y,UncertainComplex) )

        self.assertTrue( equivalent_complex( complex(1-a,-b), value(y), TOL) )
        self.assertTrue( equivalent( 1, uncertainty(y).real, TOL) )
        self.assertTrue( equivalent( 0, uncertainty(y).imag, TOL) )

        # components
        self.assertTrue( equivalent_sequence( u_component(y,x), (1,0,0,0) ) )
        self.assertTrue( equivalent_sequence( u_component(y,z), (0,0,0,0) ) )

        # same thing
        xz = ucomplex(1,(1,0))
        yz = xz - z
        self.assertTrue( equivalent_complex( value(yz), value(y), TOL) )
        self.assertTrue( equivalent( uncertainty(yz).real, uncertainty(y).real, TOL) )
        self.assertTrue( equivalent( uncertainty(yz).imag, uncertainty(y).imag, TOL) )
        
        # complex on right        
        y = z - x
        self.assertTrue( isinstance(y,UncertainComplex) )

        self.assertTrue( equivalent_complex( complex(a-1,b), value(y), TOL) )
        self.assertTrue( equivalent( 1, uncertainty(y).real, TOL) )
        self.assertTrue( equivalent( 0, uncertainty(y).imag, TOL) )

        # components
        self.assertTrue( equivalent_sequence( u_component(y,x), (-1,0,0,0) ) )
        self.assertTrue( equivalent_sequence( u_component(y,z), (0,0,0,0) ) )

        # same thing
        yz = z - xz
        self.assertTrue( equivalent_complex( value(yz), value(y), TOL) )
        self.assertTrue( equivalent( uncertainty(yz).real, uncertainty(y).real, TOL) )
        self.assertTrue( equivalent( uncertainty(yz).imag, uncertainty(y).imag, TOL) )

        # intermediate components
        # -- on the right
        a = 3
        b = 0.5
        z = complex(a,b)
        x1v = 3.1
        x2v = 4.1
        x1 = result(ureal(x1v,1) + ureal(x2v,1) )
        y = x1 - z

        ux1 = uncertainty(x1)
        self.assertTrue( equivalent_sequence( u_component(y,x1), (ux1,0,0,0) ) )

        # -- on the left
        y = z - x1
        self.assertTrue( equivalent_sequence( u_component(y,x1), (-ux1,0,0,0) ) )
      
    def test_multiplication(self):
        x = ureal(1,1)
        a = 3
        b = 0.5
        z = complex(a,b)

        y = x * z
        self.assertTrue( isinstance(y,UncertainComplex) )

        self.assertTrue( equivalent_complex( complex(a,b), value(y), TOL) )
        self.assertTrue( equivalent( a, uncertainty(y).real, TOL) )
        self.assertTrue( equivalent( b, uncertainty(y).imag, TOL) )

        # components
        self.assertTrue( equivalent_sequence( u_component(y,x), (z.real,0,z.imag,0) ) )
        self.assertTrue( equivalent_sequence( u_component(y,z), (0,0,0,0) ) )

        # same thing
        xz = ucomplex(1,(1,0))
        yz = xz * z
        self.assertTrue( equivalent_complex( value(yz), value(y), TOL) )
        self.assertTrue( equivalent( uncertainty(yz).real, uncertainty(y).real, TOL) )
        self.assertTrue( equivalent( uncertainty(yz).imag, uncertainty(y).imag, TOL) )

        self.assertTrue( equivalent_sequence( u_component(y,x), u_component(yz,xz), ) )
        
        # complex on right        
        y = z * x
        self.assertTrue( isinstance(y,UncertainComplex) )

        self.assertTrue( equivalent_complex( complex(a,b), value(y), TOL) )
        self.assertTrue( equivalent( a, uncertainty(y).real, TOL) )
        self.assertTrue( equivalent( b, uncertainty(y).imag, TOL) )

        # components
        self.assertTrue( equivalent_sequence( u_component(y,x), (z.real,0,z.imag,0) ) )
        self.assertTrue( equivalent_sequence( u_component(y,z), (0,0,0,0) ) )

        # same thing
        yz = z * xz
        self.assertTrue( equivalent_complex( value(yz), value(y), TOL) )
        self.assertTrue( equivalent( uncertainty(yz).real, uncertainty(y).real, TOL) )
        self.assertTrue( equivalent( uncertainty(yz).imag, uncertainty(y).imag, TOL) )

        self.assertTrue( equivalent_sequence( u_component(y,x), u_component(yz,xz), ) )

        # intermediate components
        # -- on the right
        a = 3
        b = 0.5
        z = complex(a,b)
        x1v = 3.1
        x2v = 4.1
        x1 = result(ureal(x1v,1) + ureal(x2v,1) )
        y = x1 * z

        ux1 = uncertainty(x1)
        dy_dx = ux1 * z
        self.assertTrue( equivalent_sequence( u_component(y,x1), (dy_dx.real,0,dy_dx.imag,0) ) )

        # -- on the left
        y = z * x1
        self.assertTrue( equivalent_sequence( u_component(y,x1), (dy_dx.real,0,dy_dx.imag,0) ) )
      
    def test_division(self):
        xv = 2
        x = ureal(xv,1)
        a = 3.0
        b = 0.5
        z = complex(a,b)

        y = x / z
        self.assertTrue( isinstance(y,UncertainComplex) )

        inv_z = 1.0/z 
        self.assertTrue( equivalent_complex( xv/z, value(y), TOL) )
        self.assertTrue( equivalent( abs(inv_z.real), uncertainty(y).real, TOL) )
        self.assertTrue( equivalent( abs(inv_z.imag), uncertainty(y).imag, TOL) )

        # components
        self.assertTrue( equivalent_sequence( u_component(y,x), (inv_z.real,0,inv_z.imag,0) ) )
        self.assertTrue( equivalent_sequence( u_component(y,z), (0,0,0,0) ) )

        # same thing
        xz = ucomplex(2,(1,0))
        yz = xz / z
        self.assertTrue( equivalent_complex( value(yz), value(y), TOL) )
        self.assertTrue( equivalent( uncertainty(yz).real, uncertainty(y).real, TOL) )
        self.assertTrue( equivalent( uncertainty(yz).imag, uncertainty(y).imag, TOL) )

        self.assertTrue( equivalent_sequence( u_component(y,x), u_component(yz,xz), ) )

        # complex on right        
        y = z / x
        self.assertTrue( isinstance(y,UncertainComplex) )

        self.assertTrue( equivalent_complex( z/xv, value(y), TOL) )
        self.assertTrue( equivalent( abs(z.real)/xv**2, uncertainty(y).real, TOL) )
        self.assertTrue( equivalent( abs(z.imag)/xv**2, uncertainty(y).imag, TOL) )

        # components
        self.assertTrue( equivalent_sequence( u_component(y,x), (-z.real/xv**2,0,-z.imag/xv**2,0) ) )
        self.assertTrue( equivalent_sequence( u_component(y,z), (0,0,0,0) ) )

        # same thing
        yz = z / xz
        self.assertTrue( equivalent_complex( value(yz), value(y), TOL) )
        self.assertTrue( equivalent( uncertainty(yz).real, uncertainty(y).real, TOL) )
        self.assertTrue( equivalent( uncertainty(yz).imag, uncertainty(y).imag, TOL) )

        self.assertTrue( equivalent_sequence( u_component(y,x), u_component(yz,xz), ) )

        # intermediate components
        # -- on the right
        a = 3
        b = 0.5
        z = complex(a,b)
        x1v = 3.1
        x2v = 4.1
        x1 = result(ureal(x1v,1) + ureal(x2v,1) )
        y = x1 / z

        ux1 = uncertainty(x1)
        dy_dx = ux1 * 1.0/z
        self.assertTrue( equivalent_sequence( u_component(y,x1), (dy_dx.real,0,dy_dx.imag,0) ) )

        # -- on the left
        y = z / x1
        dy_dx = -ux1 * z / value(x1)**2
        self.assertTrue( equivalent_sequence( u_component(y,x1), (dy_dx.real,0,dy_dx.imag,0) ) )

    def test_power(self):
        # complex on the right ----------------------------------
        xv = 3.1
        xu = 1
        x = ureal(xv,xu)
        xc = ucomplex(complex(xv,0),(xu,0))
        a = 1.2
        b = 0.3
        z = complex(a,b)

        y = x ** z
        yc = xc ** z
        
        self.assertTrue( isinstance(y,UncertainComplex) )
        self.assertTrue( equivalent_complex( xv ** z, value(y), TOL) )
        self.assertTrue( equivalent_complex( value(yc), value(y), TOL) )
        
        # components
        dy_dx =z * xv**(z-1) 
        self.assertTrue( equivalent_sequence( u_component(y,x), (dy_dx.real,0,dy_dx.imag,0) ) )
        self.assertTrue( equivalent_sequence( u_component(y,z), (0,0,0,0) ) )

        # make sure that results are equivalent to full complex calculation
        self.assertTrue( equivalent( uncertainty(yc).real, uncertainty(y).real, TOL) )
        self.assertTrue( equivalent( uncertainty(yc).imag, uncertainty(y).imag, TOL) )

        self.assertTrue( equivalent_sequence( u_component(y,x), u_component(yc,xc), ) )

        # Different numbers
        xv = -3.1
        xu = 1
        x = ureal(xv,xu)
        xc = ucomplex(complex(xv,0),(xu,0))
        a = -0.2
        b = -1.3
        z = complex(a,b)

        y = x ** z
        yc = xc ** z
        
        self.assertTrue( isinstance(y,UncertainComplex) )
        self.assertTrue( equivalent_complex( xv ** z, value(y), TOL) )
        self.assertTrue( equivalent_complex( value(yc), value(y), TOL) )

        # components
        dy_dx = z * xv**(z-1)
        self.assertTrue( equivalent_sequence( u_component(y,x), (dy_dx.real,0,dy_dx.imag,0) ) )
        self.assertTrue( equivalent_sequence( u_component(y,z), (0,0,0,0) ) )

        # make sure that results are equivalent to full complex calculation
        self.assertTrue( equivalent( uncertainty(yc).real, uncertainty(y).real, TOL) )
        self.assertTrue( equivalent( uncertainty(yc).imag, uncertainty(y).imag, TOL) )

        self.assertTrue( equivalent_sequence( u_component(y,x), u_component(yc,xc), ) )

        # complex on the left ----------------------------------
        xv = 3.1
        xu = 1
        x = ureal(xv,xu)
        xc = ucomplex(complex(xv,0),(xu,0))
        a = 1.2
        b = 0.3
        z = complex(a,b)

        y = z ** x
        yc = z ** xc
        yv = z ** xv
        
        self.assertTrue( isinstance(y,UncertainComplex) )
        self.assertTrue( equivalent_complex( yv, value(y), TOL) )
        self.assertTrue( equivalent_complex( value(yc), value(y), TOL) )

        # components
        dy_dx = yv * cmath.log(z) if z != 0 else 0
        self.assertTrue( equivalent_sequence( u_component(y,x), (dy_dx.real,0,dy_dx.imag,0) ) )
        self.assertTrue( equivalent_sequence( u_component(y,z), (0,0,0,0) ) )

        # make sure that results are equivalent to full complex calculation
        self.assertTrue( equivalent( uncertainty(yc).real, uncertainty(y).real, TOL) )
        self.assertTrue( equivalent( uncertainty(yc).imag, uncertainty(y).imag, TOL) )

        self.assertTrue( equivalent_sequence( u_component(y,x), u_component(yc,xc), ) )

        # different numbers
        xv = -3.1
        xu = 1
        x = ureal(xv,xu)
        xc = ucomplex(complex(xv,0),(xu,0))
        a = -0.2
        b = -1.3
        z = complex(a,b)

        y = z ** x
        yc = z ** xc
        yv = z ** xv
        
        self.assertTrue( isinstance(y,UncertainComplex) )
        self.assertTrue( equivalent_complex( yv, value(y), TOL) )
        self.assertTrue( equivalent_complex( value(yc), value(y), TOL) )

        # components
        dy_dx = yv * cmath.log(z) if z != 0 else 0
        self.assertTrue( equivalent_sequence( u_component(y,x), (dy_dx.real,0,dy_dx.imag,0) ) )
        self.assertTrue( equivalent_sequence( u_component(y,z), (0,0,0,0) ) )

        # make sure that results are equivalent to full complex calculation
        self.assertTrue( equivalent( uncertainty(yc).real, uncertainty(y).real, TOL) )
        self.assertTrue( equivalent( uncertainty(yc).imag, uncertainty(y).imag, TOL) )
        
        self.assertTrue( equivalent_sequence( u_component(y,x), u_component(yc,xc), ) )

        # Special case z==0 here
        xv = 3.1
        xu = 1
        x = ureal(xv,xu)
        xc = ucomplex(complex(xv,0),(xu,0))
        a = 0
        b = 0
        z = complex(a,b)

        y = z ** x
        yc = z ** xc
        yv = z ** xv
        
        self.assertTrue( isinstance(y,UncertainComplex) )
        self.assertTrue( equivalent_complex( yv, value(y), TOL) )
        self.assertTrue( equivalent_complex( value(yc), value(y), TOL) )

        # components
        dy_dx = yv * cmath.log(z) if z != 0 else 0
        self.assertTrue( equivalent_sequence( u_component(y,x), (dy_dx.real,0,dy_dx.imag,0) ) )
        self.assertTrue( equivalent_sequence( u_component(y,z), (0,0,0,0) ) )

        # make sure that results are equivalent to full complex calculation
        self.assertTrue( equivalent( uncertainty(yc).real, uncertainty(y).real, TOL) )
        self.assertTrue( equivalent( uncertainty(yc).imag, uncertainty(y).imag, TOL) )

        self.assertTrue( equivalent_sequence( u_component(y,x), u_component(yc,xc), ) )

        # intermediate components
        # -- on the right
        a = 3
        b = 0.5
        z = complex(a,b)
        x1v = 3.1
        x2v = 4.1
        x1 = result(ureal(x1v,1) + ureal(x2v,1) )
        y = x1 ** z
        yv = value(y)

        ux1 = uncertainty(x1)
        dy_dx = z * value(x1)**(z-1) * ux1
        self.assertTrue( equivalent_sequence( u_component(y,x1), (dy_dx.real,0,dy_dx.imag,0) ) )

        # -- on the left
        y = z ** x1
        yv = value(y)
        dy_dx = yv * cmath.log(z) * ux1 if z != 0 else 0
        self.assertTrue( equivalent_sequence( u_component(y,x1), (dy_dx.real,0,dy_dx.imag,0) ) )

#-----------------------------------------------------
class TestGetSetCorrelation(unittest.TestCase):

    """
    tests the extended forms of get_correlation and
    set_correlation, where uncertain complex arguments
    are given.
    """
    
    def test(self):
        x1 = ureal(1,1)
        z1 = ucomplex(1,(1,1),independent=False)
        z2 = ucomplex(1,(1,1),independent=False)
        r = (.1,.2,.3,.4)
        set_correlation(r,z1,z2)
        check_r = get_correlation(z1,z2)
        self.assertTrue( equivalent_sequence(r,check_r) )

        r_t = (.1,.3,.2,.4)
        check_r = get_correlation(z2,z1)
        self.assertTrue( equivalent_sequence(r_t,check_r) )

        self.assertRaises(TypeError,set_correlation,r,z1,x1)
        self.assertRaises(TypeError,set_correlation,r,x1,z1)

    def test_with_mixed_unumbers(self):
        x1 = ureal(1,2,independent=False)
        z1 = ucomplex(1,.5,independent=False)
        
        r1 = .1
        set_correlation(r1,z1.real,x1)
        check_r = get_correlation(z1,x1)
        self.assertTrue( equivalent_sequence([r1,0,0,0],check_r) )
        
        r2 = -.3
        set_correlation(r2,z1.imag,x1)
        check_r = get_correlation(z1,x1)
        self.assertTrue( equivalent_sequence([r1,0,r2,0],check_r) )
        
        check_r = get_correlation(x1,z1)
        self.assertTrue( equivalent_sequence([r1,r2,0,0],check_r) )

        check_r = get_correlation(1.0,z1)
        self.assertTrue( equivalent_sequence([0]*4,check_r) )

        check_r = get_correlation(1.0+7j)
        self.assertTrue( equivalent(0,check_r) )

        check_r = get_correlation(1.0+7j,z1)
        self.assertTrue( equivalent_sequence([0]*4,check_r) )

        check_r = get_correlation(z1,1.0+7j)
        self.assertTrue( equivalent_sequence([0]*4,check_r) )

        check_r = get_correlation(1.0+7j,x1)
        self.assertTrue( equivalent_sequence([0]*4,check_r) )

        check_r = get_correlation(x1,1.0+7j)
        self.assertTrue( equivalent_sequence([0]*4,check_r) )
      
        check_r = get_correlation(7,x1)
        self.assertTrue( equivalent(0,check_r) )

        check_r = get_correlation(x1,1.0)
        self.assertTrue( equivalent(0,check_r) )
        
#-----------------------------------------------------
class TestGetCovariance(unittest.TestCase):

    """
    tests get_covariance
    """
    
    def cv(self,u1,u2,r):
        return (
            u1[0]*r[0]*u2[0],
            u1[0]*r[1]*u2[1],
            u1[1]*r[2]*u2[0],
            u1[1]*r[3]*u2[1],
        )
    def test(self):
        ux1 = [0.25,0]
        uz1 = [0.33,1.5]
        uz2 = [.2,.4]
        x1 = ureal(1,ux1[0],independent=False)
        z1 = ucomplex(1,uz1,independent=False)
        z2 = ucomplex(1,uz2,independent=False)
        r = (.1,.2,.3,.4)
        
        set_correlation(r,z1,z2)
        check_r = get_covariance(z1,z2)
        self.assertTrue( equivalent_sequence(self.cv(uz1,uz2,r),check_r) )

        r_t = (.1,.3,.2,.4)
        check_r = get_covariance(z2,z1)
        self.assertTrue( equivalent_sequence(self.cv(uz2,uz1,r_t),check_r) )

        rx = [-.3,0,.7,0]
        set_correlation(rx[0],z1.real,x1)
        set_correlation(rx[2],z1.imag,x1)
        check_r = get_covariance(z1,x1)
        self.assertTrue( equivalent_sequence(self.cv(uz1,ux1,rx),check_r) )

        check_r = get_covariance(x1,z1)
        rx_t = [-.3,.7,0,0]
        self.assertTrue( equivalent_sequence(self.cv(ux1,uz1,rx_t),check_r) )

    def test_with_mixed_unumbers(self):
        ux1 = [2,0]
        x1 = ureal(1,ux1[0],independent=False)
        uz1 = [.5,.5]
        z1 = ucomplex(1,uz1,independent=False)
        
        r1 = .1
        set_correlation(r1,z1.real,x1)
        check_r = get_covariance(z1,x1)
        self.assertTrue( equivalent_sequence(self.cv(uz1,ux1,[r1,0,0,0]),check_r) )
        
        r2 = -.3
        set_correlation(r2,z1.imag,x1)
        check_r = get_covariance(z1,x1)
        self.assertTrue( equivalent_sequence(self.cv(uz1,ux1,[r1,0,r2,0]),check_r) )
        
        check_r = get_covariance(x1,z1)
        self.assertTrue( equivalent_sequence(self.cv(ux1,uz1,[r1,r2,0,0]),check_r) )

        check_r = get_covariance(1.0,z1)
        self.assertTrue( equivalent_sequence([0]*4,check_r) )

        check_r = get_covariance(1.0+7j)
        self.assertTrue( equivalent(0,check_r) )

        check_r = get_covariance(1.0+7j,z1)
        self.assertTrue( equivalent_sequence([0]*4,check_r) )

        check_r = get_covariance(z1,1.0+7j)
        self.assertTrue( equivalent_sequence([0]*4,check_r) )

        check_r = get_covariance(1.0+7j,x1)
        self.assertTrue( equivalent_sequence([0]*4,check_r) )

        check_r = get_covariance(x1,1.0+7j)
        self.assertTrue( equivalent_sequence([0]*4,check_r) )
      
        check_r = get_covariance(7,x1)
        self.assertTrue( equivalent(0,check_r) )

        check_r = get_covariance(x1,1.0)
        self.assertTrue( equivalent(0,check_r) )
 
#-----------------------------------------------------
class TestStringRepresentations(unittest.TestCase):
 
    def test(self):
        z = 1+0j
        u = 0.1 
        df = 'inf' 

        uc = ucomplex(z, u)
        s = re.search(r'ucomplex\(\((.*)\), u=\[(.*)\], r=(.*), df=(.*)\)', repr(uc))
        
        self.assertEqual(s.group(1),"{0.real:.16g}{0.imag:+.16g}j".format(z))
        self.assertEqual(s.group(2),"{0!r},{0!r}".format(u))
        self.assertEqual(s.group(3),repr(0.))
        self.assertEqual(s.group(4),"inf")
  
    def test_strange_cases(self):
        z = UncertainComplex._elementary(
            complex(inf,inf),
            inf,inf,
            None,
            inf,
            None,True
        )
        self.assertEqual( str(z), '(inf(inf)+inf(inf)j)')
        self.assertEqual( repr(z), 'ucomplex((inf+infj), u=[inf,inf], r=nan, df=inf)')
        
        z = UncertainComplex._elementary(
            complex(inf,inf),
            inf,inf,
            None,
            nan,
            None,True
        )
        self.assertEqual( str(z), '(inf(inf)+inf(inf)j)')
        self.assertEqual( repr(z), 'ucomplex((inf+infj), u=[inf,inf], r=nan, df=nan)')

        z = UncertainComplex._elementary(
            complex(inf,inf),
            inf,nan,
            None,
            nan,
            None,True
        )
        self.assertEqual( str(z), '(inf(inf)+inf(nan)j)')
        self.assertEqual( repr(z), 'ucomplex((inf+infj), u=[inf,nan], r=nan, df=nan)')

        z = UncertainComplex._elementary(
            complex(inf,inf),
            nan,nan,
            None,
            nan,
            None,True
        )
        self.assertEqual( str(z), '(inf(nan)+inf(nan)j)')
        self.assertEqual( repr(z), 'ucomplex((inf+infj), u=[nan,nan], r=nan, df=nan)')

        z = UncertainComplex._elementary(
            complex(inf,nan),
            nan,nan,
            None,
            nan,
            None,True
        )
        self.assertEqual( str(z), '(inf(nan)+nan(nan)j)')
        self.assertEqual( repr(z), 'ucomplex((inf+nanj), u=[nan,nan], r=nan, df=nan)')

        z = UncertainComplex._elementary(
            complex(nan,nan),
            nan,nan,
            None,
            nan,
            None,True
        )
        self.assertEqual( str(z), '(nan(nan)+nan(nan)j)')
        self.assertEqual( repr(z), 'ucomplex((nan+nanj), u=[nan,nan], r=nan, df=nan)')

        z = UncertainComplex._elementary(
            complex(1,1),
            0.1,0.1,
            None,
            nan,
            None,True
        )
        self.assertEqual( str(z), '(1.00(10)+1.00(10)j)')
        self.assertEqual( repr(z), 'ucomplex((1+1j), u=[0.1,0.1], r=0.0, df=nan)')

#-----------------------------------------------------
class TestMisc(unittest.TestCase):
 
    def test_constant(self):
        z = 1+0j
        uc = constant(z)
        self.assertTrue( _is_uncertain_complex_constant(uc) )
        self.assertRaises(TypeError,_is_uncertain_complex_constant,z)
        
        # Setting u=0 makes a constant (at the moment, should this be changed?)
        un = ucomplex(1j,0)
        self.assertTrue( _is_uncertain_complex_constant(un) )
        
    def test_nonzero(self):
        un = ucomplex(1+0j,1)
        self.assertTrue( bool(un) is True )
        un = ucomplex(1j,1)
        self.assertTrue( bool(un) is True )
        un = ucomplex(0+0j,1)
        self.assertTrue( bool(un) is not True )
  
    def test_variance_and_v(self):
        v_cv = (2,1,1,3)
        uz = ucomplex(0,v_cv)
        self.assertTrue(
            equivalent_sequence(
                v_cv,
                uz.v
            )
        )
        self.assertTrue(
            equivalent_sequence(
                v_cv,
                variance(uz)
            )
        )

    def test_intermediate_constant(self):
        self.assertTrue( equivalent_complex( result(1+3j), 1+3j) )
        self.assertTrue( isinstance( result(1+3j), numbers.Complex) )
        if sys.version_info[0] > 2:
            x = self.assertWarns( RuntimeWarning, result,(1+3j),label='error')
            x = self.assertWarns( RuntimeWarning, result,1,label='error')
        
#============================================================================
if(__name__== '__main__'):

    unittest.main()    # Runs all test methods starting with 'test'


import unittest
import math
try:
    xrange  # Python 2
except NameError:
    xrange = range

try:
    from itertools import izip  # Python 2
except ImportError:
    izip = zip
    xrange = range

import numpy as np

from GTC import function

from GTC.core import (
    ureal,
    ucomplex,
    value,
    uncertainty,
    variance,
    dof
)
from GTC.linear_algebra import matmul, uarray

from testing_tools import *

TOL = 1E-13 
          
#---------------------------------------------------------
class StdDataSets(object):
    """
    See section 5 in:
    'Design and us of reference data sets for testing scientific software'
    M. G. Cox and P. M. Harris, Analytica Chemica Acta 380 (1999) 339-351.
    
    """
    
    def __init__(self,mu,h,q,n):
        self._mu = mu
        self._h = h
        self._q = q
        self._n = n

    def seq(self,k=1):
        self._k = k
        
        N = self._n
        a = np.array( xrange(-N,N+1) ) * self._h
        q = self._q ** self._k
        
        return (self._mu + q) - a

    def mean(self):
        return self._mu + (self._q ** self._k)

    def std(self):
        N = self._n
        return self._h * math.sqrt((N + 0.5)*(N+1)/3.0)
        
#-----------------------------------------------------
class TestComplexToSeq(unittest.TestCase):

    """
    tests the conversion between complex and matrix representations
    """

    def test(self):
        z = 1 + 2j
        zm = function.complex_to_seq(z)
        self.assertTrue( equivalent_sequence((z.real,-z.imag,z.imag,z.real),zm ) )

        z = -1.1 - 4.5j
        zm = (z.real,-z.imag,z.imag,z.real)
        self.assertTrue( equivalent_complex(z,function.seq_to_complex(zm) ) )

        z = 1 
        zm = function.complex_to_seq(z)
        self.assertTrue( equivalent_sequence((z.real,-z.imag,z.imag,z.real),zm ) )

        z = 0 - 4.5j
        zm = (z.real,-z.imag,z.imag,z.real)
        self.assertTrue( equivalent_complex(z,function.seq_to_complex(zm) ) )

        self.assertRaises(RuntimeError,function.seq_to_complex,"hell")
        self.assertRaises(RuntimeError,function.seq_to_complex,(1,2,3,1))   # ill-conditioned
        self.assertRaises(RuntimeError,function.seq_to_complex,(1,2,2,1))   # ill-conditioned
        self.assertRaises(RuntimeError,function.seq_to_complex,(2,-2,2,1))  # ill-conditioned
        self.assertRaises(RuntimeError,function.seq_to_complex,(1,2,3,4,5))
        self.assertRaises(RuntimeError,function.seq_to_complex,(1,2,5))
        self.assertRaises(RuntimeError,function.seq_to_complex,z)

        zm = np.array(zm)   
        self.assertTrue( zm.shape == (4,) )
        self.assertRaises(RuntimeError,function.seq_to_complex,zm)

#-----------------------------------------------------
class TestFunctionFunctions(unittest.TestCase):
    """
    Test functions in `function` module:
        - mean

    """
    def setUp(self):
        self.x = [ureal(23, 2), ureal(26, 1), ureal(20, 3), ureal(25, 1), ureal(28, 2), ureal(24, 2)]
        self.y = [ureal(25, 2), ureal(24, 4), ureal(23, 1), ureal(25, 2), ureal(19, 6), ureal(27, 1)]
        self.xa = uarray(self.x)
        self.ya = uarray(self.y)

        self.xc = [ucomplex(23+2j, 2), ucomplex(26+1j, 1), ucomplex(20+0j, 3),
                   ucomplex(25+3j, 1), ucomplex(22+7j, 2), ucomplex(23+3j, 2)]
        self.yc = [ucomplex(25+1j, 2), ucomplex(24+4j, 4), ucomplex(23+2j, 1),
                   ucomplex(25+1j, 2), ucomplex(26+2j, 1), ucomplex(18+8j, 5)]
        self.xca = uarray(self.xc)
        self.yca = uarray(self.yc)
        
    def test_mean(self):
        # generator argument (iterable)
        m = function.mean( i for i in range(10) )
        self.assertTrue( equivalent(m,4.5) )
        
        # sequence argument 
        m = function.mean( range(10) )
        self.assertTrue( equivalent(m,4.5) )
        
        # anything else makes no sense
        self.assertRaises(RuntimeError,function.mean,3)
 
    def test_sum_ndarray(self):
        # 1D array
        xlist = [ureal(i, i*0.1) for i in range(100)]
        xarray = uarray(xlist)
        self.assertTrue(xarray.shape == (100,))

        b = 0
        for x in xlist:
            b += x
        for a in [function.sum(xarray)]:
            self.assertTrue(equivalent(value(a), b.x))
            self.assertTrue(equivalent(uncertainty(a), b.u))

        # 3D array
        xlist = [[[ureal(i*j*k, i*j*k*0.1) for k in range(1, 5)] for j in range(7, 10)] for i in range(3, 9)]
        xarray = uarray(xlist)
        self.assertTrue(xarray.shape == (6, 3, 4))

        axis_none = 0.0
        axis_0 = [[0.0 for i in range(4)] for j in range(3)]
        axis_1 = [[0.0 for i in range(4)] for j in range(6)]
        axis_2 = [[0.0 for i in range(3)] for j in range(6)]
        for i in range(6):
            for j in range(3):
                for k in range(4):
                    _value = xlist[i][j][k]
                    axis_none += _value
                    axis_0[j][k] += _value
                    axis_1[i][k] += _value
                    axis_2[i][j] += _value

        # axis=None
        for a in [function.sum(xarray)]:
            self.assertTrue(equivalent(value(a), axis_none.x))
            self.assertTrue(equivalent(uncertainty(a), axis_none.u))

        # axis=0
        m, n = len(axis_0), len(axis_0[0])
        for a in [function.sum(xarray, axis=0)]:
            self.assertTrue(a.shape == (m, n))
            for j in range(m):
                for k in range(n):
                    self.assertTrue(equivalent(a[j, k].x, axis_0[j][k].x))
                    self.assertTrue(equivalent(a[j, k].u, axis_0[j][k].u))

        # axis=1
        m, n = len(axis_1), len(axis_1[0])
        for a in [function.sum(xarray, axis=1)]:
            self.assertTrue(a.shape == (m, n))
            for i in range(m):
                for k in range(n):
                    self.assertTrue(equivalent(a[i, k].x, axis_1[i][k].x))
                    self.assertTrue(equivalent(a[i, k].u, axis_1[i][k].u))

        # axis=2
        m, n = len(axis_2), len(axis_2[0])
        for a in [function.sum(xarray, axis=2)]:
            self.assertTrue(a.shape == (m, n))
            for i in range(m):
                for j in range(n):
                    self.assertTrue(equivalent(a[i, j].x, axis_2[i][j].x))
                    self.assertTrue(equivalent(a[i, j].u, axis_2[i][j].u)) 
                    
    def test_mean_ndarray(self):
        for j, (x, xa) in enumerate([(self.x, self.xa), (self.xc, self.xca)]):
            ave = 0.0
            for val in x:
                ave += val
            ave = ave/float(len(x))

        for m in [function.mean(xa)]:
            if j == 0:
                self.assertTrue(equivalent(value(m), ave.x))
                self.assertTrue(equivalent(uncertainty(m), ave.u))
            else:
                self.assertTrue(equivalent_complex(value(m), ave.x))
                self.assertTrue(equivalent(uncertainty(m).real, ave.u.real))
                self.assertTrue(equivalent(uncertainty(m).imag, ave.u.imag))

        xa = xa.reshape(2, 3)

        for m in [function.mean(xa)]:
            if j == 0:
                self.assertTrue(equivalent(value(m), ave.x))
                self.assertTrue(equivalent(uncertainty(m), ave.u))
            else:
                self.assertTrue(equivalent_complex(value(m), ave.x))
                self.assertTrue(equivalent(uncertainty(m).real, ave.u.real))
                self.assertTrue(equivalent(uncertainty(m).imag, ave.u.imag))

        for m in [function.mean(xa, axis=0)]:
            aves = [(x[0] + x[3])/2.0, (x[1] + x[4])/2.0, (x[2] + x[5])/2.0]
            for idx in range(3):
                if j == 0:
                    self.assertTrue(equivalent(m[idx].x, aves[idx].x))
                    self.assertTrue(equivalent(m[idx].u, aves[idx].u))
                else:
                    self.assertTrue(equivalent_complex(m[idx].x, aves[idx].x))
                    self.assertTrue(equivalent(m[idx].u.real, aves[idx].u.real))
                    self.assertTrue(equivalent(m[idx].u.imag, aves[idx].u.imag))

        for m in [function.mean(xa, axis=1)]:
            aves = [(x[0] + x[1] + x[2])/3.0, (x[3] + x[4] + x[5])/3.0]
            for idx in range(2):
                if j == 0:
                    self.assertTrue(equivalent(m[idx].x, aves[idx].x))
                    self.assertTrue(equivalent(m[idx].u, aves[idx].u))
                else:
                    self.assertTrue(equivalent_complex(m[idx].x, aves[idx].x))
                    self.assertTrue(equivalent(m[idx].u.real, aves[idx].u.real))
                    self.assertTrue(equivalent(m[idx].u.imag, aves[idx].u.imag))
#============================================================================
if(__name__== '__main__'):

    unittest.main()    # Runs all test methods starting with 'test'


"""
Least-squares regression
------------------------

:func:`~type_b.line_fit` implements an ordinary least-squares 
straight-line regression calculation that accepts uncertain 
real numbers for the independent and dependent variables.

:func:`~type_b.line_fit_wls` implements a weighted least-squares 
straight-line regression calculation. It accepts uncertain 
real numbers for the independent and dependent variables.
It is also possible to specify weights for the regression.

:func:`~type_b.line_fit_wtls` implements a total least-squares 
algorithm for a straight-line fitting that can perform a 
weighted least-squares regression when both `y` and `x` data  
are uncertain real numbers, it also handles correlation 
between (x,y) data pairs.

Probability distributions: real-valued problems 
-----------------------------------------------
    
    Functions that convert the half-width   
    of a one-dimensional distribution to  
    a standard uncertainty:
    
    *   :func:`uniform`
    *   :func:`triangular`
    *   :func:`u_shaped` 
    *   :func:`arcsine` 

Probability distributions: complex-valued problems
--------------------------------------------------
    
    Functions that convert information
    about two-dimensional distributions into
    standard uncertainties:
    
    *   :func:`uniform_ring`
    *   :func:`uniform_disk`
    *   :func:`unknown_phase_product`

A table of distributions
------------------------

    The mapping :obj:`distribution` is provided so that 
    the functions above can be selected by name. 
    For example, ::

        >>> a = 1.5
        >>> ureal( 1, type_b.distribution['gaussian'](a) )
        ureal(1.0,1.5,inf)
        >>> ureal( 1, type_b.distribution['uniform'](a) )
        ureal(1.0,0.8660254037844387,inf)
        >>> ureal( 1, type_b.distribution['arcsine'](a) )
        ureal(1.0,1.0606601717798212,inf)

    Keys to :obj:`.distribution` are (case-sensitive):
    
    *   `gaussian`
    *   `uniform`
    *   `triangular`
    *   `arcsine`
    *   `u_shaped`
    *   `uniform_ring`
    *   `uniform_disk`
    
Module contents
---------------

"""
from __future__ import division

import sys
import math

import numpy as np

try:  # Python 2
    import __builtin__ as builtins
    from collections import Iterable
    from itertools import izip
except ImportError:
    import builtins
    from collections.abc import Iterable
    izip = zip
    xrange = range
    
from GTC import (
    inf ,
    EPSILON,
    is_sequence

)

from GTC.named_tuples import InterceptSlope
from GTC.vector import scale_vector

from GTC.lib import (
    UncertainReal,
    value,
    value_seq
)

__all__ = (
    'uniform',
    'triangular',
    'u_shaped',
    'arcsine',
    'uniform_ring',
    'uniform_disk',
    'unknown_phase_product',
    'distribution',
    'mean',
    'line_fit','line_fit_wls','line_fit_wtls',
    'LineFitOLS', 'LineFitWLS', 'LineFitWTLS'
)

HALF_PI = math.pi / 2.0
MAX = sys.float_info.max 

#-----------------------------------------------------------------------------------------
def mean(seq,*args,**kwargs):
    """Return the arithmetic mean of data in ``seq``

    :arg seq: a sequence, :class:`~numpy.ndarray`, or iterable, of numbers or uncertain numbers
    :arg args: optional arguments when ``seq`` is an :class:`~numpy.ndarray`
    :arg kwargs: optional keyword arguments when ``seq`` is an :class:`~numpy.ndarray`
    
    An uncertain number is returned if ``seq`` contains uncertain numbers.
            
    """
    if is_sequence(seq):
        assert not args
        assert not kwargs
        mu = sum(seq)/len(seq)
        
    elif isinstance(seq,np.ndarray):
        mu = np.asarray(seq).mean(*args, **kwargs)
        
    elif isinstance(seq,Iterable):
        assert not args
        assert not kwargs
        count = 0
        total = 0
        for i in seq:
            total += i
            count += 1
        mu = total/count
        
    else:
        raise RuntimeError(
            "{!r} is not iterable".format(seq)
        )
    
    # If `seq` has uncertain number elements then `mu` will 
    # be an uncertain number.     
    return mu
#-----------------------------------------------------------------------------------------
class LineFit(object):
    
    """
    Base class for the results of regression to a line.
    
    .. versionadded:: 1.2
    """
    
    def __init__(self,a,b,ssr,N):
        self._a_b = InterceptSlope(a,b)
        self._ssr = ssr
        self._N = N
        
    def __repr__(self):
        return """{}(
  a={!r},
  b={!r},
  ssr={},
  N={}
)""".format(
            self.__class__.__name__,
            self._a_b[0],
            self._a_b[1],
            self._ssr,
            self._N
        )

    @property
    def a_b(self):
        """Return the intercept ``a`` and slope ``b`` as a tuple of uncertain numbers
        """
        return self._a_b

    @property
    def intercept(self):
        """Return the intercept as an uncertain number.
        """
        return self._a_b[0]

    @property
    def slope(self):
        """Return the slope as an uncertain number.
        """
        return self._a_b[1]

    @property
    def ssr(self):
        """Sum of the squared residuals
        
        The sum of the squared deviations between values 
        predicted by the model and the actual data.
        
        If weights are used during the fit, the squares of 
        weighted deviations are summed.
        
        """
        return self._ssr  

    @property
    def N(self):
        """The number of points in the sample"""
        return self._N

    def __str__(self):
        a, b = self.a_b
        return '''
  Intercept: {!s}
  Slope: {!s}
  Correlation: {:.2G}
  Sum of the squared residuals: {}
  Number of points: {}
'''.format(
    a,
    b,
    a.get_correlation(b),
    self._ssr,
    self.N
)
 
#-----------------------------------------------------------------------------------------
class LineFitOLS(LineFit):
    
    """
    Class to hold results from an ordinary linear regression to data.
    
    .. versionadded:: 1.2
    """
    
    def __init__(self,a,b,ssr,N):
        LineFit.__init__(self,a,b,ssr,N)

    def __str__(self):
        header = '''
Ordinary Least-Squares Results:
'''
        return header + LineFit.__str__(self)
        
    def x_from_y(self,yseq,x_label=None):
        """Estimate the stimulus ``x`` corresponding to the responses in ``yseq``

        :arg yseq: a sequence of further observations of ``y``
        :arg x_label: a label for the return uncertain number `x` 

        The items in ``yseq`` must be uncertain real numbers.
        
        .. note::
            When ``x_label`` is defined, the uncertain number returned will be 
            declared an intermediate result (using :func:`~.result`)
        
        """
        a, b = self._a_b
        
        y = mean( yseq ) 
        
        if x_label is None:
            return (y - a)/b
        else:
            return result( (y - a)/b, label=x_label )

    def y_from_x(self,x,y_label=None):
        """Return an uncertain number ``y`` that predicts the response to ``x``

        :arg x: an uncertain real number
        :arg y_label: a label for the return uncertain number `y` 

        This is a prediction of a single future response ``y`` to a stimulus ``x``
        
        .. note::
            When ``y_label`` is defined, the uncertain number returned will be 
            declared an intermediate result (using :func:`~.result`)
        
        """
        a, b = self._a_b   
                          
        if y_label is None:
            return a + b*x
        else:
            return result( a + b*x, label=y_label )
        
#-----------------------------------------------------------------------------------------

class LineFitWLS(LineFitOLS):
    
    """
    This object holds results from a weighted LS linear regression to data.
    
    .. versionadded:: 1.2
    """
    
    def __init__(self,a,b,ssr,N):
        LineFitOLS.__init__(self,a,b,ssr,N)

    def __str__(self):
        header = '''
Weighted Least-Squares Results:
'''
        return header + LineFit.__str__(self)
        

#-----------------------------------------------------------------------------------------
class LineFitWTLS(LineFit):
    
    """
    This object holds results from a TLS linear regression to data.
    
    .. versionadded:: 1.2
    """
    
    def __init__(self,a,b,ssr,N):
        LineFit.__init__(self,a,b,ssr,N)

    def __str__(self):
        header = '''
Weighted Total Least-Squares Results:
'''
        return header + LineFit.__str__(self)


#--------------------------------------------------------------------
#
def line_fit(x,y):
    """Least-squares fit intercept and slope 
    
    .. versionadded:: 1.2
    
    :arg x:     sequence of independent variable data 
    :arg y:     sequence of dependent variable data

    :rtype:     a :class:`~type_b.LineFitOLS`
    
    ``y`` must be a sequence of uncertain real numbers.

    Performs an ordinary least-squares regression. 
    
    .. note::

        Uncertainty in the parameter estimates is found
        by propagation *through* the regression
        formulae. This does **not** take residuals into account.
        
        The function :func:`type_a.line_fit` performs a regression 
        analysis that evaluates uncertainty in 
        the parameter estimates using the residuals.
        
        If appropriate, the results from both type-A and type-B 
        analyses can be merged (see :func:`type_a.merge`).
        
    **Example**::

        >>> a0 =10
        >>> b0 = -3
        >>> u0 = .2

        >>> x = [ float(x_i) for x_i in xrange(10) ]
        >>> y = [ ureal(b0*x_i + a0,u0) for x_i in x ]

        >>> a,b = tb.line_fit(x,y).a_b
        >>> a
        ureal(10.0,0.1175507627290...,inf)
        >>> b
        ureal(-3.0,0.02201927530252...,inf)
        
    """  
    S = len(x) 
    
    S_x = sum( x ) 
    S_y = sum( y )

    k = S_x / S
    t = [ x_i - k for x_i in x ]

    S_tt = sum( t_i*t_i for t_i in t )
    
    b = sum( t_i*y_i/S_tt for t_i,y_i in izip(t,y) )
    a = (S_y - b*S_x)/S

    if not isinstance(a, UncertainReal):
        raise ValueError('"y" must be a sequence of uncertain real numbers. '
                         'You may want to use type_a.line_fit instead.')

    # The sum of squared residuals is now calculated but not used
    float_a = value(a)
    float_b = value(b)
    
    f2 = lambda x_i,y_i: (
        (y_i - float_a - float_b*x_i)
    )**2 
    
    ssr =  math.fsum( 
        f2( value(x_i), value(y_i) ) 
            for x_i,y_i in izip(x,y) 
    )

    return LineFitOLS(a,b,ssr,S)
 
#--------------------------------------------------------------------
#
def line_fit_wls(x,y,u_y=None):
    """-> Weighted least-squares linear regression
    
    .. versionadded:: 1.2
    
    :arg x:     sequence of independent variable data 
    :arg y:     sequence of dependent variable data
    :arg u_y:   sequence of uncertainties in ``y``

    :rtype:    :class:`~type_b.LineFitWLS` 
    
    ``y`` must be a sequence of uncertain real numbers.

    Performs a weighted least-squares regression. 

    Weights are calculated from the uncertainty of 
    the ``y`` elements unless the sequence ``u_y`` 
    is provided. 
    
    .. note::

        The uncertainty in the parameter estimates is found
        by propagation of uncertainty *through* the regression
        formulae. This does **not** take account of the residuals.
        
        The function :func:`type_a.line_fit_wls` can be used to 
        carry out a regression analysis that obtains uncertainty in 
        the parameter estimates due to the residuals.
        
        If necessary, the results of both type-A and type-B 
        analyses can be merged (see :func:`type_a.merge`).
        
    **Example**::

        >>> x = [1,2,3,4,5,6]
        >>> y = [3.2, 4.3, 7.6, 8.6, 11.7, 12.8]
        >>> u_y = [0.5,0.5,0.5,1.0,1.0,1.0]
        >>> y = [ ureal(y_i,u_y_i) for y_i, u_y_i in zip(y,u_y) ]
        
        >>> fit = type_b.line_fit_wls(x,y)
        >>> a, b = fit.a_b
        >>> a
        ureal(0.8852320675105...,0.5297081435088...,inf)
        >>> b
        ureal(2.0569620253164...,0.1778920167412...,inf)
        
    """
    if u_y is None:
        v = [ y_i.v for y_i in y ]
        u = [ math.sqrt(v_i) for v_i in v ]
    else:
        v = [ u_y_i**2 for u_y_i in u_y ]
        u = u_y     
        
    S = sum( 1.0/v_i for v_i in v)
    S_x = sum( x_i/v_i for x_i,v_i in izip(x,v) )    
    S_y = sum( y_i/v_i for y_i,v_i in izip(y,v) )

    k = S_x / S
    t = [ (x_i - k)/u_i for x_i,u_i in izip(x,u) ]

    S_tt = sum( t_i*t_i for t_i in t )

    b = sum( t_i*y_i/u_i/S_tt for t_i,y_i,u_i in izip(t,y,u) )
    a = (S_y - b*S_x)/S

    if not isinstance(a, UncertainReal):
        raise ValueError('"y" must be a sequence of uncertain real numbers. '
                         'You may want to use type_a.line_fit_wls instead.')

    # The sum of squared residuals is now calculated but not used
    float_a = value(a)
    float_b = value(b)
    
    f2 = lambda x_i,y_i,u_i: (
        (y_i - float_a - float_b*x_i)/u_i
    )**2 
    
    ssr =  math.fsum( 
        f2( value(x_i), value(y_i), u_i ) 
            for x_i,y_i,u_i in izip(x,y,u) 
    )

    return LineFitWLS(a,b,ssr,len(x))

#--------------------------------------------------------------------
ZEPS = 1E-10
def _dbrent(ax,bx,cx,fn,tol=math.sqrt(EPSILON)):
    """
    Minimise fn() and return x, fn(x) and df_dx(x), all floats
    
    `fn` must be a univariate function of an uncertain real that returns
    an uncertain real number.

    `ax`, `bx` and `cx` must be floats. `bx` must be between `ax` and `cx`
    and fn(bx) must be less than both fn(ax) and fn(cx).

    `context` - a GTC context
    
    `tol` - the fractional precision

    See also Numerical Recipes in C, 2nd ed, Section 10.3
    
    """
    ITMAX = 100

    deriv = lambda y,x: y.sensitivity(x)
    ureal = lambda x,u: UncertainReal._elementary(
            x,u,inf,None,True
    )    
    e = 0.0 # The distance moved on the step before last
    
    a = ax if ax < cx else cx
    b = ax if ax > cx else cx
    
    assert a <= bx and bx <= b, "Invalid initial values in _dbrent"
        
    # if fn( ureal(bx,1)) > fn( ureal(a,1)) or fn(ureal(bx,1)) > fn(ureal(b,1)):
        # assert False

    x = w = v = bx

    _u_ = ureal(x,1.0)    
    fn_u = fn( _u_ )
    
    fw = fv = fx = value(fn_u)
    dw = dv = dx = deriv(fn_u,_u_)

    # The routine keeps track of `a` and `b`, which bracket the minimum,
    # `x` is the point with the least function value found so far,
    # `w` is the point with the second least value, `v` is the previous
    # value of `w`, `u` is the point at which the function was most 
    # recently evaluated.
    
    for i in xrange(ITMAX):
        
        xm = 0.5*(a + b)
        tol1 = tol*abs(x) + ZEPS
        tol2 = 2.0*tol1

        if abs(x - xm) <= ( tol2 - 0.5*(b - a) ):
            return x, fx, dx
        
        if abs(e) > tol1:
            # initialise the d's to be out of bracket 
            d1 = 2.0*(b - a)
            d2 = d1
            
            # Secant method
            if dw != dx: d1 = (w - x)*dx/(dx - dw)  
            if dv != dx: d2 = (v - x)*dx/(dx - dv)  
            
            # Choose one estimate.
            # Insist that it be within the bracket 
            # and on the side pointed to by the derivative at `x` 
            u1 = x + d1
            u2 = x + d2
            OK1 = (a - u1)*(u1 - b) > 0.0 and dx*d1 <= 0.0
            OK2 = (a - u2)*(u2 - b) > 0.0 and dx*d2 <= 0.0
            
            olde, e = e, d

            if OK1 or OK2:
                if OK1 and OK2:
                    d = d1 if abs(d1) < abs(d2) else d2
                elif OK1:
                    d = d1
                else:
                    d = d2
                    
                if abs(d) <= abs(0.5 * olde):
                    u = x + d
                    if (u - a < tol2) or (b - u < tol2):
                        d = math.copysign(tol1,xm-x)
                else:
                    # choose segment by the sign of the derivative
                    e = a - x if dx >= 0.0 else b - x
                    d = 0.5*e
            else:
                e = a - x if dx >= 0.0 else b - x
                d = 0.5*e
                
        else:
            e = a - x if dx >= 0.0 else b - x
            d = 0.5*e
            
        if abs(d) >= tol1:
            u = x + d
            
            _u_ = ureal(u,1.0)
            fn_u = fn(_u_)
            
            fu = value(fn_u)
            du = deriv(fn_u,_u_)
            
        else:
            # Smallest step possible 
            u = x + math.copysign(tol1,d)
            
            _u_ = ureal(u,1.0)
            fn_u = fn(_u_)
            
            fu = value(fn_u)
            du = deriv(fn_u,_u_)
            
            # If the minimum sized step downhill 
            # goes up, then we are done!
            if fu > fx:
                return u, fu, du
                
        assert a <= u and u <= b, (a,u,b)   # invariant
    
        if fu <= fx:
            # Found a new best point 
            
            # Update the bracket on one side so that the previous 
            # `x` value is now the limit and the new `x` value is 
            # contained. 
            if u >= x:
                a = x
            else:
                b = x
                
            v, fv, dv = w, fw, dw
            w, fw, dw = x, fx, dx
            x, fx, dx = u, fu, du
            
            assert a <= x and x <= b, (a,x,b)   # invariant
            
        else:
            # The point `x` has not been bettered
            
            # `x` does not change, but `u` was inside the 
            # bracket so we can tighten the noose.
            if u < x:
                a = u
            else:
                b = u
             
            # `w` is the second best point and `v` is the 3rd best 
            if (fu <= fw) or (x == w):
                v, fv, dv = w, fw, dw
                w, fw, dw = u, fu, du
                
            elif (fu < fv) or (v == x) or (v == w):            
                v, fv, dv = u, fu, du
   
            assert a <= x and x <= b, (a,x,b)   # invariant
            
        assert fx <= fw and fx <= fv # invariant
        
    raise RuntimeError('Exceeded iteration limit in `_dbrent`')

#--------------------------------------------------------------------
def _arrays(sin_a,cos_a,sin_2a,cos_2a,x,y,u2_x,u2_y,cov):
    """Returns the set of arrays needed for the Chi-sq calculation

    Equations reference: M Krystek and M Anton,
    Meas. Sci. Technol. 22 (2011) 035101 (9pp)

    Note, this utility function can work equally with real numbers
    or uncertain real numbers. This allows us to us it in the
    type_a module and checking routines too.
    
    """
    sin_a_2 = sin_a**2
    cos_a_2 = cos_a**2
    two_sin_cos_a = 2.0*sin_a*cos_a

    # Note an alternative (perhaps more stable numerically is the following
    # it will need cos_2a and sin_2a to be passed in as arguments too.
    # However, it fails one of our test cases because an element goes to zero!
    
    # eqn(53)
    g_k = [
        (u2_x_i + u2_y_i)/2.0 - (u2_x_i - u2_y_i)*cos_2a/2.0 - 2.0*cov_i*sin_2a
            for u2_x_i, u2_y_i, cov_i in izip(u2_x,u2_y,cov)
    ]

##    # eqn(32)
##    g_k = [
##        u2_x_i*sin_a_2 + u2_y_i*cos_a_2 - two_sin_cos_a*cov_i
##            for u2_x_i, u2_y_i, cov_i in izip(u2_x,u2_y,cov)
##    ]

    N = len(g_k)

    fix_div_by_zero = lambda y,x: y/x if x != 0 else MAX

    # eqn(33), but without sqrt
    u2 = 1.0/(
        sum( fix_div_by_zero(1.0,g_k_i) for g_k_i in g_k ) / N
    )
    # eqn(34)
    w_k = [ fix_div_by_zero(u2,g_k_i) for g_k_i in g_k ]

    # Eqns (35,36,43)
    x_bar = sum(
        w_k_i*x_i for w_k_i,x_i in izip(w_k,x)
    ) / N
    
    y_bar = sum(
        w_k_i*y_i for w_k_i,y_i in izip(w_k,y)
    ) / N

    p_hat = y_bar*cos_a - x_bar*sin_a    
    
    # eqn(31)
    v_k = [ y_i*cos_a - x_i*sin_a - p_hat for x_i,y_i in izip(x,y) ]  

    return v_k,u2_x,u2_y,g_k,u2,x_bar,y_bar,p_hat

#--------------------------------------------------------------------
class ChiSq(object):

    """
    A callable object representing Chi-squared as a function of alpha.
    
    ``x`` and ``y`` are lists of uncertain real numbers used to 
    initialise the object. 
    
    Equations reference: M Krystek and M Anton,
    Meas. Sci. Technol. 22 (2011) 035101 (9pp)
    
    .. versionadded:: 1.2
    """
    
    def __init__(self,x,y,u_x,u_y,r_xy):
        self.x_u = x
        self.y_u = y
        
        self.x = [ x_i.x for x_i in x ]
        self.y = [ y_i.x for y_i in y ]

        if u_x is not None:
            self.u2_x = [ u_i**2 for u_i in u_x ]
            self.u2_y = [ u_i**2 for u_i in u_y ]
            self.cov = [ u_x_i*u_y_i*r_i
                for u_x_i,u_y_i,r_i in izip(u_x,u_y,r_xy) 
            ]  
        else:    
            self.u2_x = [ x_i.v for x_i in x ]
            self.u2_y = [ y_i.v for y_i in y ]
        
            self.cov = [ x_i.get_covariance(y_i)
                for x_i,y_i in izip(x,y)
            ]
        
    #--------------------------------------------------------------------
    def arrays(self,alpha):
        """Return v_k,g_k      
        """
        sin_a = alpha._sin()
        sin_2a = (2.0*alpha)._sin()
        cos_a = alpha._cos()
        cos_2a = (2.0*alpha)._cos()

        v_k,u2_x,u2_y,g_k,u2,x_bar,y_bar,p_hat = _arrays(
            sin_a,cos_a,sin_2a,cos_2a,self.x,self.y,self.u2_x,self.u2_y,self.cov
        )

        return v_k, g_k

    #--------------------------------------------------------------------
    def p_hat(self,alpha):
        """Eqn (43)"""
        sin_a = alpha._sin()
        sin_2a = (2.0*alpha)._sin()
        cos_a = alpha._cos()
        cos_2a = (2.0*alpha)._cos()

        v_k,u2_x,u2_y,g_k,u2,x_bar,y_bar,p_hat = _arrays(
            sin_a,cos_a,sin_2a,cos_2a,self.x_u,self.y_u,self.u2_x,self.u2_y,self.cov
        )

        return p_hat
        
    #--------------------------------------------------------------------
    def __call__(self,alpha):
        """Returns ChiSq(alpha)"""
        v_k,g_k = self.arrays(alpha)
            
        # Eqn (30)
        chi_2 = sum(
            v_k_i**2/g_k_i for v_k_i,g_k_i in izip(v_k,g_k)
        )
        
        return chi_2

#--------------------------------------------------------------------
class dChiSq_dalpha(object):

    """
    Callable object representing the derivative of Chi-squared wrt alpha. 

    .. versionadded:: 1.2
    
    Equations reference: M Krystek and M Anton,
    Meas. Sci. Technol. 22 (2011) 035101 (9pp)

    Some definitions (equation numbers):
        v_k :  (31)
        v_ka : (51)
        v_kaa : (52)
        g_k : (53)
        g_ka : 54        
    """
    
    def __init__(self,x,y,u_x,u_y,r_xy):
        self.x = x
        self.y = y

        if u_x is not None:
            self.u2_x = [ u_i**2 for u_i in u_x ]
            self.u2_y = [ u_i**2 for u_i in u_y ]
            self.cov = [ u_x_i*u_y_i*r_i
                for u_x_i,u_y_i,r_i in izip(u_x,u_y,r_xy) 
            ]  
        else:
            self.u2_x = [ x_i.v for x_i in x ]
            self.u2_y = [ y_i.v for y_i in y ]
            self.cov = [ x_i.get_covariance(y_i)
                for x_i,y_i in izip(x,y) 
            ]      
        
    #--------------------------------------------------------------------
    def arrays(self,alpha):
        """Return v_k, v_ka, g_k, g_ka (eqns: 31, 51, 53, 54)
        """
        sin_a = alpha._sin()
        sin_2a = (2.0*alpha)._sin()
        cos_a = alpha._cos()
        cos_2a = (2.0*alpha)._cos()

        v_k,u2_x,u2_y,g_k,u2,x_bar,y_bar,p_hat = _arrays(
            sin_a,cos_a,sin_2a,cos_2a,self.x,self.y,self.u2_x,self.u2_y,self.cov
        )

        # d(v_k)_d(alpha)
        v_ka = [ -y_i*sin_a - x_i*cos_a for x_i,y_i in izip(self.x,self.y) ]
        
        # d(g_k)_d(alpha)
        g_ka = [ sin_2a*(u2_x_i-u2_y_i) - 2.0*cov_i*cos_2a
            for u2_x_i,u2_y_i,cov_i in izip(u2_x,u2_y,self.cov) ]
        
        return v_k, v_ka, g_k, g_ka

        
    #--------------------------------------------------------------------
    def __call__(self,alpha):
        """Returns d(ChiSq)_d(alpha)( alpha ) 
        
        `v_k` (31); `v_ka` (51); `v_kaa` (52)
        `g_k` (53); `g_ka` (54); 
        """
        v_k, v_ka, g_k, g_ka = self.arrays(alpha)       
        
        # Eqn (56)
        dy_dx = sum(
           (2.0*v_k_i*g_k_i*v_ka_i - v_k_i**2*g_ka_i)/g_k_i**2
                for v_k_i,v_ka_i,g_k_i,g_ka_i
                    in izip(v_k, v_ka, g_k, g_ka)
        )
        
        return dy_dx
     
#--------------------------------------------------------------------
def line_fit_wtls(x,y,u_x=None,u_y=None,a_b=None,r_xy=None):
    """Perform straight-line regression with uncertainty in ``x`` and ``y``

    .. versionadded:: 1.2
    
    :arg x: list of uncertain real numbers for the independent variable
    :arg y: list of uncertain real numbers for the dependent variable
    :arg u_x: a sequence of uncertainties for the ``x`` data
    :arg u_y: a sequence of uncertainties for the ``y`` data
    :arg a_b: a pair of initial estimates for the intercept and slope
    :arg r_xy: correlation between x-y pairs [default: 0]

    Returns a :class:`~type_b.LineFitWTLS` object

    The elements of ``x`` and ``y`` must be uncertain numbers
    with non-zero uncertainties. If specified, the optional arguments 
    ``u_x`` and ``u_y`` will be used uncertainties to weight 
    the data for the regression, otherwise the uncertainties of
    the uncertain numbers in the sequences are used.
    
    The optional argument ``a_b`` can be used to provide a pair 
    of initial estimates for the intercept and slope. Otherwise, 
    initial estimates will be obtained by calling `line_fit_wls`.
    
    Implements a Weighted Total Least Squares algorithm
    that allows for correlation between x-y pairs. See reference: 
    
    M Krystek and M Anton, *Meas. Sci. Technol.* **22** (2011) 035101 (9pp)
        
    **Example**::

        # Pearson-York test data
        # see, e.g., Lybanon, M. in Am. J. Phys 52 (1), January 1984 
        >>> xin=[0.0,0.9,1.8,2.6,3.3,4.4,5.2,6.1,6.5,7.4]
        >>> wx=[1000.0,1000.0,500.0,800.0,200.0,80.0,60.0,20.0,1.8,1.0]
        >>> yin=[5.9,5.4,4.4,4.6,3.5,3.7,2.8,2.8,2.4,1.5]
        >>> wy=[1.0,1.8,4.0,8.0,20.0,20.0,70.0,70.0,100.0,500.0]

        # Convert weights to standard uncertainties 
        >>> uxin=[1./math.sqrt(wx_i) for wx_i in wx ]
        >>> uyin=[1./math.sqrt(wy_i) for wy_i in wy ]

        # Define uncertain numbers
        >>> x = [ ureal(xin_i,uxin_i) for xin_i,uxin_i in zip(xin,uxin) ]
        >>> y = [ ureal(yin_i,uyin_i) for yin_i,uyin_i in zip(yin,uyin) ]

        # TLS returns uncertain numbers
        >>> a,b = type_b.line_fit_wtls(x,y).a_b
        >>> a
        ureal(5.47991018...,0.29193349...,inf)
        >>> b
        ureal(-0.48053339...,0.057616740...,inf)

    """
    ureal = lambda x,u: UncertainReal._elementary(
            x,u,inf,None,True
    )    

    if (u_x is not None or u_y is not None):
        if (u_x is None or u_y is None):
            raise RuntimeError(
                "You must supply ``u_x`` and ``u_y``"
            )
        elif (r_xy is None):
            # default value will be uncorrelated
            r_xy = [0] * len(u_x)
            
    for x_i,y_i in izip(x,y):
        assert isinstance(x_i,UncertainReal), 'uncertain real required'
        assert isinstance(y_i,UncertainReal), 'uncertain real required'

    if a_b is None:
        a_b = line_fit_wls(x, y, u_y).a_b

    a0 = value(a_b[0])
    b0 = value(a_b[1])
    
    # initial value for `alpha`
    alpha0 = math.atan(b0)

    # chi_sq(alpha0) -> chisquared
    chi_sq = ChiSq(x,y,u_x,u_y,r_xy)   

    # Search for the minimum chi-squared wrt alpha
    x1 = alpha0 - HALF_PI
    x2 = alpha0 + HALF_PI

    # `brent` requires three points that bracket the minimum.
    # the `x1`, `alpha0`, `x2` parameters should be real, 
    # but `data` will return an uncertain number 
    # and expects an uncertain number argument.
    #
    # Returns x, fn(x) and df_dx(x), all floats
    alpha1,fn_alpha1,df_alpha1 = _dbrent(x1,alpha0,x2,chi_sq)

    # dChiSq_a( alpha ) will return dChiSq_dalpha(`alpha`)
    # dChiSq_a(alpha0) -> 1st partial derivative of chisquared at alpha0
    dChiSq_a = dChiSq_dalpha(x,y,u_x,u_y,r_xy)   
    
    # Need the partial derivative of dChiSq_a wrt alpha 
    alpha = ureal(alpha1,1) 
    F_alpha = dChiSq_a( alpha ) 
    dalpha_dF = -1.0/F_alpha.sensitivity(alpha)

    # Now we define `alpha` with sensitivity to the ``x`` and ``y`` data,
    # via the object ``F_alpha``, which represents the 1st partial derivative
    # of chi-squared at alpha1 (ideally zero, but really only close to the root).
    F_alpha = dChiSq_a( UncertainReal._constant(alpha1) )
    alpha = UncertainReal(
        alpha1
    ,   scale_vector(F_alpha._u_components,dalpha_dF)
    ,   scale_vector(F_alpha._d_components,dalpha_dF)
    ,   scale_vector(F_alpha._i_components,dalpha_dF)
    )

    # The sensitivity of p_hat to the x and y data is via 
    # `alpha`, `x_bar` and `y_bar` in eqn (43)
    p_hat = chi_sq.p_hat( alpha )

    # Note we have reversed the definitions of `a` and `b` here
    b = alpha._tan()
    a = p_hat/alpha._cos()

    N = len(x)

    ssr = chi_sq( UncertainReal._constant(alpha1) ).x

    return LineFitWTLS(a,b,ssr,N)
    
#---------------------------------------------------------------------------
_root_2 = math.sqrt(2.0)
_root_3 = math.sqrt(3.0)
_root_6 = math.sqrt(6.0)

#---------------------------------------------------------------------------
def uniform(a):
    """Return the standard uncertainty for a uniform distribution. 

    :arg float a: the half-width

    **Example**::

        >>> x = ureal(1,type_b.uniform(1))
        >>> x
        ureal(1.0,0.5773502691896258,inf)
    
    """
    return a/_root_3

#---------------------------------------------------------------------------
def triangular(a):
    """Return the standard uncertainty for a triangular distribution. 
    
    :arg float a: the half-width 
    
    **Example**::

        >>> x = ureal(1,type_b.triangular(1))
        >>> x
        ureal(1.0,0.4082482904638631,inf)
        
    """
    return a/_root_6

#---------------------------------------------------------------------------
def arcsine(a):
    """Return the standard uncertainty for an arcsine distribution. 

    :arg float a: the half-width 
    
    **Example**::

        >>> x = ureal(1,type_b.arcsine(1))
        >>> x
        ureal(1.0,0.7071067811865475,inf)

    """
    return a/_root_2

# Aliases for the arcsine function
u_shaped = arcsine

#---------------------------------------------------------------------------
def uniform_ring(a):
    """Return the standard uncertainty for a uniform ring
    
    :arg float a: the radius
    
    Convert the radius of a uniform ring distribution ``a`` 
    to a standard uncertainty

    See reference: B D Hall, *Metrologia* **48** (2011) 324-332
    
    **Example**::

        >>> z = ucomplex( 0, type_b.uniform_ring(1) )
        >>> z
        ucomplex((0+0j), u=[0.7071067811865475,0.7071067811865475], r=0.0, df=inf)
        
    """
    return arcsine(a)

#---------------------------------------------------------------------------
def uniform_disk(a):
    """Return the standard uncertainty for a uniform disk 
    
    :arg float a: the radius
    
    Convert the radius of a uniform disk distribution ``a`` 
    to a standard uncertainty.
    
    See reference: B D Hall, *Metrologia* **48** (2011) 324-332

    **Example**::

        >>> z = ucomplex( 0, type_b.uniform_disk(1) )
        >>> z
        ucomplex((0+0j), u=[0.5,0.5], r=0.0, df=inf)
        
    """
    return a / 2.0

# #---------------------------------------------------------------------------
# # 
# def uncertain_ring(a_u_r):
    # """Return the standard uncertainty for an uncertain ring 
        
    # :arg a_u_r: the (estimated) radius with a standard uncertainty 
    # :type a_u_r: pair of float
    
    # Convert a radius estimate ``a``, with a standard
    # uncertainty ``u_r``, into a standard uncertainty.

    # See reference: B D Hall, *Metrologia* **48** (2011) 324-332

    # **Example**::

        # >>> estimate = (1,0.1)
        # >>> z = ucomplex( 0, type_b.uncertain_ring( estimate ) )
        # >>> z
        # ucomplex((0+0j), u=[0.714142842854285,0.714142842854285], r=0.0, df=inf)

    # .. note::
    
        # This function is deprecated. 
        
    # """
    # a, u_r = a_u_r
    # return math.sqrt( a**2/2.0 + u_r**2 )

#---------------------------------------------------------------------------
def unknown_phase_product(u1,u2):
    """Return the standard uncertainty for a product when phases are unknown

    :arg u1: the standard uncertainty of the first multiplicand
    :arg u2: the standard uncertainty of the second multiplicand
    
    Obtains the standard uncertainty associated
    with a complex product when estimates have unknown phase.

    The arguments ``u1`` and ``u2`` are the standard  
    uncertainties associated with each multiplicand.  
    
    See reference: B D Hall, *Metrologia* **48** (2011) 324-332

    **Example**::
    
        # X = Gamma1 * Gamma2
        >>> X = ucomplex( 0, type_b.unknown_phase_product(.1,.1) )
        >>> X
        ucomplex((0+0j), u=[0.014142135623730954,0.014142135623730954], r=0.0, df=inf)

    """
    return _root_2 * u1 * u2
   
#---------------------------------------------------------------------------
#
distribution = dict(
    gaussian = lambda x: x,
    uniform = uniform,
    triangular = triangular,
    arcsine = arcsine,
    u_shaped = u_shaped,
    uniform_ring = uniform_ring,
    uniform_disk = uniform_disk,
    # 'uncertain_ring': uncertain_ring
)
"""Maps ``distribution[name]`` -> `function` for standard uncertainty

    **Example**::
        >>> type_b.distribution['arcsine'](1.5)
        1.06066017177982...
"""
   
#============================================================================
if __name__ == "__main__":
    import doctest
    from GTC import *
    doctest.testmod()



"""
Copyright (c) 2018, Measurement Standards Laboratory of New Zealand.

"""

__all__ = (
    'Leaf',
    'Node'
)

#----------------------------------------------------------------------------
class Node(object):
    
    """
    A `Node` holds information about an intermediate uncertain real number
    """

    __slots__ = [
        'uid',
        'label',
        'u',
        'df',
        'complex',
        '__weakref__'
    ]
 
    def __init__(self,uid,label,u,df):    
        self.uid = uid
        self.label = label
        self.u = u
        self.df = df
 
#----------------------------------------------------------------------------
class Leaf(Node):

    """
    A `Leaf` holds information about an elementary uncertain real number
    """
    
    __slots__ = [
        'independent',
        'correlation',
        'ensemble'
    ]
    
    def __init__(self,uid,label,u,df,independent=True):
        Node.__init__(self,uid,label,u,df)
    
        self.independent = independent
        if not independent:
            self.correlation = {uid: 1.0}
            self.ensemble = set()
        
    
    